"""
This module contains functions to assist in the construction of URIs for views.
"""

__author__      = "Graham Klyne (GK@ACM.ORG)"
__copyright__   = "Copyright 2014, G. Klyne"
__license__     = "MIT (http://opensource.org/licenses/MIT)"

import logging
log = logging.getLogger(__name__)

import urllib
import re

# From RFC 3986:
gen_delims  = ":/?#[]@"
sub_delims  = "!$&'()*+,;="
unreserved  = "-._~"
# subset of above safe in query string (no "?", "&" or #")
query_safe  = re.sub('[?&#]', '', gen_delims + sub_delims + unreserved)

def build_dict(*param_dicts, **param_dict):
    merged_param_dict = param_dict.copy()
    for d in param_dicts:
        # log.info("param_dicts %r"%(d,))
        merged_param_dict.update(d)
    return merged_param_dict

def uri_params(*param_dicts, **param_dict):
    """
    Construct a URI parameter string from the supplied dictionary values 
    (or values which are convertible to a dictionary using `dict()`).
    """
    uri_param_dict = build_dict(*param_dicts, **param_dict)
    uri_param_str = ""
    next_sep      = "?"        
    for pnam in uri_param_dict:
        pval = uri_param_dict[pnam]
        if pval is not None:
            uri_param_str += next_sep + pnam + "=" + urllib.quote(pval, query_safe)
            next_sep = "&"
    return uri_param_str

def uri_base(uri):
    """
    Construct a base URI from the supplied base URI by removing any parameters and/or fragments.
    """
    base_uri = uri.split("#", 1)[0]
    base_uri = base_uri.split("?", 1)[0]
    return base_uri

def uri_with_params(base_uri, *param_dicts, **param_dict):
    """
    Construct a URI from the supplied base URI (with any parameters and/or fragment removed)
    and URI paramneters created using the supplied dictionary values.
    """
    return uri_base(base_uri) + uri_params(*param_dicts, **param_dict)

def continuation_params(*param_dicts, **param_dict):
    """
    Return URI parameters from the supplied dictionary specifically needed for a continuation 
    URI, ignoring all others.  These are the parameters which, in conjunction with a base URI, 
    represent application state.  Parameters not included here are transient in their effect.
    """
    uri_param_dict = build_dict(*param_dicts, **param_dict)
    return (
        { 'continuation_url': uri_param_dict.get('continuation_url') or None
        , 'search':           uri_param_dict.get('search_for')       or 
                              uri_param_dict.get('search')           or None
        })

# End.



"""
This module contains definitions intended to reflect the site-wide data
used by all Annalist installations.

Test cases should use values returned by this module so that additions to 
the site data can be updated here, in just one place.
"""

__author__      = "Graham Klyne (GK@ACM.ORG)"
__copyright__   = "Copyright 2014, G. Klyne"
__license__     = "MIT (http://opensource.org/licenses/MIT)"

import os
import urlparse

import logging
log = logging.getLogger(__name__)

# from django.conf                import settings
# from django.http                import QueryDict
# from django.utils.http          import urlquote, urlunquote
# from django.core.urlresolvers   import resolve, reverse

# from annalist.util              import valid_id
# from annalist.identifiers       import RDF, RDFS, ANNAL
# from annalist                   import layout
# from annalist                   import message

# from tests import (
#     TestHost, TestHostUri, TestBasePath, TestBaseUri, TestBaseDir
#     )

# from entity_testsitedata            import (
#     get_site_types, get_site_types_sorted,
#     get_site_lists, get_site_lists_sorted,
#     get_site_list_types, get_site_list_types_sorted,
#     get_site_views, get_site_views_sorted,
#     get_site_field_groups, get_site_field_groups_sorted, 
#     get_site_fields, get_site_fields_sorted, 
#     get_site_field_types, get_site_field_types_sorted, 
#     )


#   -----------------------------------------------------------------------------
#
#   ----- Site-wide entities defined
#
#   -----------------------------------------------------------------------------

#   ----- Types -----

site_types = (
    [ "_initial_values"
    , "_field"
    , "_group"
    , "_list"
    , "_type"
    , "_user"
    , "_view"
    , "BibEntry_type"
    , "Default_type"
    , "Enum_bib_type"
    , "Enum_list_type"
    , "Enum_render_type"
    ])

def get_site_types_sorted():
    return site_types[1:]

def get_site_types():
    return set(site_types[1:])

#   ----- Lists -----

site_lists = (
    [ "_initial_values"
    , "BibEntry_list"
    , "Default_list"
    , "Default_list_all"
    , "Field_group_list"
    , "Field_list"
    , "List_list"
    , "Type_list"
    , "User_list"
    , "View_list"
    ])

def get_site_lists_sorted():
    return site_lists[1:]

def get_site_lists():
    return set(site_lists[1:])

#   ----- List types -----

site_list_types = (
    [ "Grid"
    , "List"
    ])

def get_site_list_types_sorted():
    return site_list_types

def get_site_list_types():
    return set(site_list_types)

#   ----- Views -----

site_views = (
    [ "_initial_values"
    , "BibEntry_view"
    , "Default_view"
    , "Field_group_view"
    , "Field_view"
    , "List_view"
    , "Type_view"
    , "User_view"
    , "View_view"
    ])

def get_site_views_sorted():
    return site_views[1:]

def get_site_views():
    return set(site_views[1:])

#   ----- Field groups -----

site_field_groups = (
    [ "_initial_values"
    , "Bib_book_group"
    , "Bib_identifier_group"
    , "Bib_journal_group"
    , "Bib_license_group"
    , "Bib_person_group"
    , "Bib_publication_group"
    , "Group_field_group"
    , "List_field_group"
    , "Type_alias_group"
    , "View_field_group"
    ])

def get_site_field_groups_sorted():
    return site_field_groups[1:]

def get_site_field_groups():
    return set(site_field_groups[1:])

#   ----- Fields -----

site_default_entity_fields = (
    [ "_initial_values"
    , "Entity_comment"
    , "Entity_id"
    , "Entity_label"
    , "Entity_type"
    ])

site_bibentry_fields = (
    [ "_initial_values"
    , "Bib_address"
    , "Bib_alternate"
    , "Bib_authors"
    , "Bib_bookentry"
    , "Bib_booktitle"
    , "Bib_chapter"
    , "Bib_description"
    , "Bib_edition"
    , "Bib_editors"
    , "Bib_eprint"
    , "Bib_firstname"
    , "Bib_howpublished"
    , "Bib_id"
    , "Bib_idanchor"
    , "Bib_identifiers"
    , "Bib_idtype"
    , "Bib_institution"
    , "Bib_journal"
    , "Bib_jurisdiction"
    , "Bib_lastname"
    , "Bib_license"
    , "Bib_month"
    , "Bib_name"
    , "Bib_note"
    , "Bib_number"
    , "Bib_organization"
    , "Bib_pages"
    , "Bib_publication_details"
    , "Bib_publisher"
    , "Bib_school"
    , "Bib_shortcode"
    , "Bib_title"
    , "Bib_type"
    , "Bib_url"
    , "Bib_volume"
    , "Bib_year"
    ])

site_field_fields = (
    [ "_initial_values"
    , "Field_comment"
    , "Field_default"
    , "Field_entity_type"
    , "Field_groupref"
    , "Field_id"
    , "Field_label"
    , "Field_missing"
    , "Field_placeholder"
    , "Field_placement"
    , "Field_property"
    , "Field_render"
    , "Field_repeat_label_add"
    , "Field_repeat_label_delete"
    , "Field_restrict"
    , "Field_type"
    , "Field_typeref"
    ])

site_group_fields = (
    [ "_initial_values"
    , "Group_comment"
    , "Group_field_sel"
    , "Group_field_placement"
    , "Group_field_property"
    , "Group_id"
    , "Group_label"
    , "Group_repeat_fields"
    ])

site_list_fields = (
    [ "_initial_values"
    , "List_choice"
    , "List_comment"
    , "List_default_type"
    , "List_default_view"
    , "List_entity_selector"
    , "List_id"
    , "List_label"
    , "List_repeat_fields"
    , "List_target_type"
    , "List_type"
    ])

site_type_fields = (
    [ "_initial_values"
    , "Type_comment"
    , "Type_id"
    , "Type_label"
    , "Type_list"
    , "Type_uri"
    , "Type_view"
    ])

site_user_fields = (
    [ "_initial_values"
    , "User_description"
    , "User_id"
    , "User_name"
    , "User_permissions"
    , "User_uri"
    ])

site_view_fields = (
    [ "_initial_values"
    # , "View_choice"
    , "View_comment"
    , "View_edit_view"
    , "View_fields"
    , "View_id"
    , "View_label"
    , "View_target_type"
    ])

site_fields = (
    [ "_initial_values" ] +
    site_bibentry_fields[1:] +
    site_default_entity_fields[1:] +
    site_field_fields[1:] +
    site_group_fields[1:] +
    site_list_fields[1:] +
    site_type_fields[1:] +
    site_user_fields[1:] +
    site_view_fields[1:] +
    [])

def get_site_fields_sorted():
    return site_fields[1:]

def get_site_fields():
    return set(site_fields[1:])

def get_site_default_entity_fields_sorted():
    return site_default_entity_fields[1:]

def get_site_default_entity_fields():
    return set(site_default_entity_fields[1:])

def get_site_view_fields_sorted():
    return site_default_entity_fields[1:] + site_view_fields[1:]

def get_site_view_fields():
    return set(get_site_view_fields_sorted())

def get_site_bibentry_fields_sorted():
    return site_bibentry_fields[1:] + site_default_entity_fields[1:]

def get_site_bibentry_fields():
    return set(get_site_bibentry_fields_sorted())

#   ----- Field render types -----

site_field_types = (
    [ "_initial_values"
    , "CheckBox"
    , "EntityId"
    , "EntityTypeId"
    , "Enum"
    , "Enum_choice"
    , "Enum_optional"
    , "Field"
    , "Identifier"
    , "List"
    , "Markdown"
    , "Placement"
    , "RepeatGroup"
    , "RepeatGroupRow"
    , "Slug"
    , "Text"
    , "Textarea"
    , "TokenSet"
    , "Type"
    , "URIImage"
    , "URILink"
    , "View"
    ])

def get_site_field_types_sorted():
    return site_field_types[1:]

def get_site_field_types():
    return set(site_field_types[1:])

# End.



"""
Tests for site data.

This module contains tests that are intended to test the validity, 
completeness and consistency of Annalist-defined site-wide data.
It assumes that other test modules confirm operation of the 
underlying web site server code.

The general pattern of these tests is to render a web page and then to 
Check that the rendered page contains appropriate contents as defined 
by the Annalist site data.

A key reason for this module's existence is to catch any regressions in the
presentation of views that have been manually checked for correctness that 
may be caused by changes to the underlying site data structures.
It is not intended to check details of web page presentation.

(The first couple of tests also check aspects of the test site data setup.)
"""

__author__      = "Graham Klyne (GK@ACM.ORG)"
__copyright__   = "Copyright 2014, G. Klyne"
__license__     = "MIT (http://opensource.org/licenses/MIT)"

import os
import unittest

import logging
log = logging.getLogger(__name__)

from django.conf                import settings
from django.contrib.auth.models import User
from django.test                import TestCase
from django.test.client         import Client

from bs4                        import BeautifulSoup

from annalist.identifiers                   import RDF, RDFS, ANNAL
from annalist.models.site                   import Site
from annalist.models.collection             import Collection
from annalist.models.recordtype             import RecordType
from annalist.models.recordlist             import RecordList
from annalist.models.recordview             import RecordView
from annalist.models.recordgroup            import RecordGroup
from annalist.models.recordfield            import RecordField

from annalist.views.fields.render_placement import get_placement_options

from AnnalistTestCase           import AnnalistTestCase
from tests                      import dict_to_str, init_annalist_test_site
from tests                      import (
    TestHost, TestHostUri, TestBasePath, TestBaseUri, TestBaseDir
    )
from entity_testutils           import (
    site_view_url,
    collection_view_url,
    collection_edit_url,
    collection_entity_list_url,
    collection_entity_view_url,
    collection_entity_edit_url,
    create_user_permissions, create_test_user
    )
from entity_testsitedata        import (
    get_site_types, get_site_types_sorted,
    get_site_lists, get_site_lists_sorted,
    get_site_list_types, get_site_list_types_sorted,
    get_site_views, get_site_views_sorted,
    get_site_field_groups, get_site_field_groups_sorted, 
    get_site_fields, get_site_fields_sorted, 
    get_site_view_fields, get_site_view_fields_sorted, 
    get_site_field_types, get_site_field_types_sorted, 
    )

#   -----------------------------------------------------------------------------
#
#   Site data tests
#
#   -----------------------------------------------------------------------------

class AnnalistSiteDataTest(AnnalistTestCase):
    """
    Tests for Annalist site data completeness and consistency
    """

    def setUp(self):
        init_annalist_test_site()
        self.testsite  = Site(TestBaseUri, TestBaseDir)
        self.coll1     = Collection.load(self.testsite, "coll1")
        self.types_expected = get_site_types_sorted() + ["type1", "type2"]
        self.lists_expected = get_site_lists_sorted() + ['list1', 'list2']
        self.views_expected = get_site_views_sorted() + ["view1", "view2"]
        self.list_types_expected   = ["Grid", "List"]
        self.render_types_expected = get_site_field_types_sorted()
        self.grouprefs_expected    = get_site_field_groups_sorted()
        self.placements_expected   = get_placement_options()

        # Login with admin permissions
        permissions=["VIEW", "CREATE", "UPDATE", "DELETE", "CONFIG", "ADMIN"]
        create_test_user(
            self.coll1, 
            "testuser", "testpass", user_permissions=permissions
            )
        self.client = Client(HTTP_HOST=TestHost)
        loggedin = self.client.login(username="testuser", password="testpass")
        self.assertTrue(loggedin)
        return

    def tearDown(self):
        return

    # --------------------------------------------------------------------------
    # Helper functions
    # --------------------------------------------------------------------------

    # Dereference URI and return BeautifulSoup object for the returned HTML
    def get_page(self, uri):
        r = self.client.get(uri)
        if r.status_code == 302:
            r = self.client.get(r['location'])
        self.assertEqual(r.status_code,   200)
        self.assertEqual(r.reason_phrase, "OK")
        s = BeautifulSoup(r.content)
        return s

    # Test named input has specified type and value
    def check_input_type_value(self, 
            s, name, 
            input_type, input_value=None, input_text=None):
        tag_name = "input"
        tag_type = input_type
        if input_type == "textarea":
            tag_name = "textarea"
            tag_type = None
        input_elem = s.form.find(attrs={"name": name})
        self.assertEqual(input_elem.name, tag_name)
        if tag_type:
            self.assertEqual(input_elem['type'], tag_type)
        if input_value is not None:
            self.assertEqual(input_elem['value'], input_value)
        if input_text is not None:
            self.assertEqual(input_elem.string, input_text)
        return

    # Convert selected Unicode characters to HTMLentity form (undoing changes wrought by BeautifulSoup)
    @staticmethod
    def html_encode(s):
        # See: http://stackoverflow.com/a/1919221/324122
        entity_map = (
            { u'\u2588': u"&block;"
            , u'\u2591': u"&blk14;"
            })
        for u, e in entity_map.iteritems():
            s = s.replace(u, e)
        return s

    # Test named input is <select> with specified options and selected value
    def check_select_field(self, s, name, options, selection):
        select_elem  = s.find("select", attrs={"name": name})
        options_here = [self.html_encode(o.string) for o in select_elem.find_all("option")]
        if options_here != options:
            log.info("options_here: %r"%(options_here,))
            log.info("options:      %r"%(options,))
        self.assertEqual(options_here, options)
        if selection is not None:
            self.assertEqual(
                select_elem.find("option", selected=True).string, 
                selection
                )
        return

    def check_row_column(self, row_data, colnum, row_expected):
        if row_expected[colnum] is not None:
            e = row_expected[colnum]
            f = (row_data
                    .find("div", class_="row")
                    .find_all("div", class_="columns")[colnum]
                    .stripped_strings.next()
                )
            self.assertEqual(e, f, "%s != %s (%r)"%(e, f, row_expected))
        return

    def check_list_row_data(self, s, trows_expected):
        trows = s.form.find_all("div", class_="tbody")
        self.assertEqual(len(trows), len(trows_expected))
        for i in range(len(trows_expected)):
            e = trows_expected[i]   # expected
            f = trows[i]            # found
            # <input class="select-box right" name="entity_select" type="checkbox" value="_list/list1"/>
            self.assertEqual(e[0], f.find("input", class_="select-box")["value"])
            for j in range(len(trows_expected[i][1])):
                self.check_row_column(trows[i], j, trows_expected[i][1])
        return

    def check_view_fields(self, s, expect_fields, expect_field_choices):
        trows = s.find_all("div", class_="select-row")
        self.assertEqual(len(trows), len(expect_fields))
        for i in range(len(trows)):
            tcols = trows[i].select("div.columns div.row div.columns")
            self.assertEqual(trows[i].div.input['type'],  "checkbox")
            self.assertEqual(trows[i].div.input['name'],  "View_fields__select_fields")
            self.assertEqual(trows[i].div.input['value'], str(i))
            self.check_select_field(
                tcols[0], "View_fields__%d__Field_id"%i, 
                expect_field_choices, expect_fields[i]
                )
        return

    # Test consistency of type / list / view / field descriptions
    def check_type_list_view(self, type_id, list_id, view_id, type_uri):
        # Read type description - check required fields are present
        type_type = RecordType.load(self.coll1, type_id, self.testsite)
        self.assertEqual(type_type["@type"],                [ANNAL.CURIE.Type])
        self.assertEqual(type_type[ANNAL.CURIE.id],         type_id)
        self.assertEqual(type_type[ANNAL.CURIE.type_id],    "_type")
        self.assertEqual(type_type[ANNAL.CURIE.uri],        type_uri)
        self.assertEqual(type_type[ANNAL.CURIE.type_list],  list_id)
        self.assertEqual(type_type[ANNAL.CURIE.type_view],  view_id)
        # Read type list description
        type_list = RecordList.load(self.coll1, list_id, self.testsite)
        self.assertEqual(type_list["@type"],                    [ANNAL.CURIE.List])
        self.assertEqual(type_list[ANNAL.CURIE.id],             list_id)
        self.assertEqual(type_list[ANNAL.CURIE.type_id],        "_list")
        self.assertEqual(type_list[ANNAL.CURIE.display_type],   "List")
        self.assertEqual(type_list[ANNAL.CURIE.default_type],   type_id)
        self.assertEqual(type_list[ANNAL.CURIE.default_view],   view_id)
        self.assertEqual(type_list[ANNAL.CURIE.record_type],    type_uri)
        self.assertIn(ANNAL.CURIE.list_entity_selector,         type_list)
        # Read type view description
        type_view = RecordView.load(self.coll1, view_id, self.testsite)
        self.assertEqual(type_view["@type"],                    [ANNAL.CURIE.View])
        self.assertEqual(type_view[ANNAL.CURIE.id],             view_id)
        self.assertEqual(type_view[ANNAL.CURIE.type_id],        "_view")
        self.assertEqual(type_view[ANNAL.CURIE.record_type],    type_uri)
        self.assertIn(ANNAL.CURIE.open_view,                    type_view)
        # Read and check fields used in list and view displays
        self.check_type_fields(type_id, type_uri, type_list[ANNAL.CURIE.list_fields])
        self.check_type_fields(type_id, type_uri, type_view[ANNAL.CURIE.view_fields])
        return

    # Test consistency of field descriptions for a given type
    def check_type_fields(self, type_id, type_uri, view_fields):
        for f in view_fields:
            field_id   = f[ANNAL.CURIE.field_id]
            view_field = RecordField.load(self.coll1, field_id, self.testsite)
            field_type = view_field[ANNAL.CURIE.field_render_type]
            try:
                self.assertEqual(view_field["@type"], [ANNAL.CURIE.Field])
                self.assertEqual(view_field[ANNAL.CURIE.id],      field_id)
                self.assertEqual(view_field[ANNAL.CURIE.type_id], "_field")
                self.assertIn(ANNAL.CURIE.property_uri,           view_field)
                self.assertIn(ANNAL.CURIE.field_render_type,      view_field)
                self.assertIn(ANNAL.CURIE.field_value_type,       view_field)
                self.assertIn(ANNAL.CURIE.field_placement,        view_field)
                self.assertIn(ANNAL.CURIE.placeholder,            view_field)
                self.assertIn(ANNAL.CURIE.field_placement,        view_field)
                if ANNAL.CURIE.field_entity_type in view_field:
                    # @@TODO: make this trequired?
                    self.assertEqual(
                        view_field[ANNAL.CURIE.field_entity_type], type_uri
                        )
                if field_type in ["RepeatGroup", "RepeatGroupRow"]:
                    # Check extra fields
                    group_id = view_field[ANNAL.CURIE.group_ref]
                    self.assertIn(ANNAL.CURIE.repeat_label_add,    view_field)
                    self.assertIn(ANNAL.CURIE.repeat_label_delete, view_field)
                    # Check field group
                    field_group = RecordGroup.load(self.coll1, group_id, self.testsite)
                    self.assertEqual(field_group["@type"], [ANNAL.CURIE.Field_group])
                    self.assertEqual(field_group[ANNAL.CURIE.id],      group_id)
                    self.assertEqual(field_group[ANNAL.CURIE.type_id], "_group")
                    self.check_type_fields(
                        "_group", ANNAL.CURIE.Field_group, field_group[ANNAL.CURIE.group_fields]
                        )
                    # field_name is present only if different from field_id
                    # self.assertIn(ANNAL.CURIE.field_name,  list_field)
                enum_types = (
                    [ "Type", "View", "List", "Field"
                    , "Enum", "Enum_optional"
                    ])
                if field_type in enum_types:
                    self.assertIn(ANNAL.CURIE.options_typeref, view_field)
            except Exception as e:
                log.warning("check_type_fields error %s, field_id %s, render_type %s"%(e, field_id, field_type))
                raise
        return

    # --------------------------------------------------------------------------
    # Test cases
    # --------------------------------------------------------------------------

    # Site front page
    def test_site_view(self):
        u = site_view_url()
        s = self.get_page(u)
        # Check displayed collections (check site setup)
        self.assertEqual(s.title.string, "Annalist data notebook test site")
        trows = s.select("form > div > div > div")
        self.assertEqual(len(trows), 5)
        for i in (1,2,3):
            tcols = trows[i].find_all("div", class_="view-value")
            colln = "coll%d"%i
            self.assertEqual(tcols[0].a.string,  colln)
            self.assertEqual(tcols[0].a['href'], collection_view_url(colln))
        return
 
    # Test collection view
    def test_collection_view(self):
        u = collection_view_url(coll_id="coll1")
        s = self.get_page(u)

        self.assertEqual(s.title.string, "Collection coll1")
        self.assertEqual(s.h3.string, "List entities with type information")
        self.check_input_type_value(s, "search_for", "text", "")
        self.check_select_field(s, "list_choice", self.lists_expected, "Default_list_all")

        thead = s.form.find("div", class_="thead").find("div", class_="row").find_all("div", class_="columns")
        self.assertEqual(thead[0].span.string, "Id")
        self.assertEqual(thead[1].span.string, "Type")
        self.assertEqual(thead[2].span.string, "Label")
        # test_default_list_all performs other relevant tests
        return

    # Test configuration view
    def test_collection_edit(self):
        u = collection_edit_url(coll_id="coll1")
        s = self.get_page(u)
        self.assertEqual(s.title.string, "Collection coll1")
        self.assertEqual(s.h3.string, "Customize collection coll1")
        types_expected = ['type1', 'type2']
        self.check_select_field(s, "typelist", types_expected, None)
        lists_expected = ['list1', 'list2']
        self.check_select_field(s, "listlist", lists_expected, None)
        views_expected = ['view1', 'view2']
        self.check_select_field(s, "viewlist", views_expected, None)
        return

    # Default list all
    # Same as test_collection_view but using different URI
    def test_default_list_all(self):
        u = collection_entity_list_url(coll_id="coll1", list_id="Default_list_all")
        s = self.get_page(u)

        self.assertEqual(s.title.string, "Collection coll1")
        self.assertEqual(s.h3.string, "List entities with type information")
        self.check_input_type_value(s, "search_for", "text", "")
        self.check_select_field(s, "list_choice", self.lists_expected, "Default_list_all")

        thead = s.form.find("div", class_="thead").find("div", class_="row").find_all("div", class_="columns")
        self.assertEqual(thead[0].span.string, "Id")
        self.assertEqual(thead[1].span.string, "Type")
        self.assertEqual(thead[2].span.string, "Label")

        trows_expected = (
            [ [ "_list/list1",    ["list1", "_list", "RecordList coll1/list1"] ]
            , [ "_list/list2",    ["list2", "_list", "RecordList coll1/list2"] ]
            , [ "_type/type1",    ["type1", "_type", "RecordType coll1/type1"] ]
            , [ "_type/type2",    ["type2", "_type", "RecordType coll1/type2"] ]
            , [ "_user/testuser", ["testuser", "_user", "Test User"] ]
            , [ "_view/view1",    ["view1", "_view", "RecordView coll1/view1"] ]
            , [ "_view/view2",    ["view2", "_view", "RecordView coll1/view2"] ]
            , [ "type1/entity1",  ["entity1", "type1", "Entity coll1/type1/entity1"] ]
            , [ "type1/entity2",  ["entity2", "type1", "Entity coll1/type1/entity2"] ]
            , [ "type1/entity3",  ["entity3", "type1", "Entity coll1/type1/entity3"] ]
            , [ "type2/entity1",  ["entity1", "type2", "Entity coll1/type2/entity1"] ]
            , [ "type2/entity2",  ["entity2", "type2", "Entity coll1/type2/entity2"] ]
            , [ "type2/entity3",  ["entity3", "type2", "Entity coll1/type2/entity3"] ]
            ])
        self.check_list_row_data(s, trows_expected)

        # @@TODO: check entity and type links
        #         check_list_row_links?  or add 3rd element to each trows_expected
        return

    # Default list
    def test_default_list(self):
        u = collection_entity_list_url(coll_id="coll1", list_id="Default_list")
        s = self.get_page(u)

        self.assertEqual(s.title.string, "Collection coll1")
        self.assertEqual(s.h3.string, "List entities")
        self.check_input_type_value(s, "search_for", "text", "")
        self.check_select_field(s, "list_choice", self.lists_expected, "Default_list")

        thead = s.form.find("div", class_="thead").find("div", class_="row").find_all("div", class_="columns")
        self.assertEqual(thead[0].span.string, "Id")
        self.assertEqual(thead[1].span.string, "Label")

        trows_expected = (
            [ [ "_list/list1",    ["list1", "RecordList coll1/list1"] ]
            , [ "_list/list2",    ["list2", "RecordList coll1/list2"] ]
            , [ "_type/type1",    ["type1", "RecordType coll1/type1"] ]
            , [ "_type/type2",    ["type2", "RecordType coll1/type2"] ]
            , [ "_user/testuser", ["testuser", "Test User"] ]
            , [ "_view/view1",    ["view1", "RecordView coll1/view1"] ]
            , [ "_view/view2",    ["view2", "RecordView coll1/view2"] ]
            , [ "type1/entity1",  ["entity1", "Entity coll1/type1/entity1"] ]
            , [ "type1/entity2",  ["entity2", "Entity coll1/type1/entity2"] ]
            , [ "type1/entity3",  ["entity3", "Entity coll1/type1/entity3"] ]
            , [ "type2/entity1",  ["entity1", "Entity coll1/type2/entity1"] ]
            , [ "type2/entity2",  ["entity2", "Entity coll1/type2/entity2"] ]
            , [ "type2/entity3",  ["entity3", "Entity coll1/type2/entity3"] ]
            ])
        self.check_list_row_data(s, trows_expected)
        return

    # --------------------------------------------------------------------------
    # Test site data for types
    # --------------------------------------------------------------------------

    # Test type / list / view / field consistency for RecordType
    def test_recordtype_type_list_view(self):
        self.check_type_list_view("_type", "Type_list", "Type_view", ANNAL.CURIE.Type)
        return

    # List types using type list
    def test_type_list(self):
        u = collection_entity_list_url(
            coll_id="coll1", list_id="Type_list", type_id="_type", scope="all"
            )
        s = self.get_page(u)

        self.assertEqual(s.title.string, "Collection coll1")
        self.assertEqual(s.h3.string, "List types")
        self.check_input_type_value(s, "search_for", "text", "")
        self.check_select_field(s, "list_choice", self.lists_expected, "Type_list")

        thead = (
            s.form.find("div", class_="thead")
                  .find("div", class_="row")
                  .find_all("div", class_="columns")
            )
        self.assertEqual(thead[0].span.string, "Id")
        self.assertEqual(thead[1].span.string, "Label")

        trows_expected = (
            [ [ "_type/_field",           ["_field",           "Field description type"] ]
            , [ "_type/_group",           ["_group",           "Field group type"] ]
            , [ "_type/_initial_values",  ["_initial_values",  None] ]
            , [ "_type/_list",            ["_list",            "List description type"] ]
            , [ "_type/_type",            ["_type",            "Type description type"] ]
            , [ "_type/_user",            ["_user",            "User permissions type"] ]
            , [ "_type/_view",            ["_view",            "View description type"] ]
            , [ "_type/BibEntry_type",    ["BibEntry_type",    "Bibliographic record type"] ]
            , [ "_type/Default_type",     ["Default_type",     "Default record type"] ]
            , [ "_type/Enum_bib_type",    ["Enum_bib_type",    "Bibliographic entry type"] ]
            , [ "_type/Enum_list_type",   ["Enum_list_type",   "List display type"] ]
            , [ "_type/Enum_render_type", ["Enum_render_type", "Field render type"] ]
            , [ "_type/type1",            ["type1",            "RecordType coll1/type1"] ]
            , [ "_type/type2",            ["type2",            "RecordType coll1/type2"] ]
            ])
        self.check_list_row_data(s, trows_expected)
        return

    # Create/edit type using type view
    def test_type_edit_new(self):
        u = collection_entity_edit_url(
            coll_id="coll1", type_id="_type",
            view_id="Type_view", action="new"
            )
        s = self.get_page(u)
        self.check_input_type_value(s, "entity_id", "text", None)
        self.check_input_type_value(s, "Type_label", "text", None)
        self.check_input_type_value(s, "Type_comment", "textarea", None)
        self.check_input_type_value(s, "Type_uri", "text", None)
        self.check_select_field(s, "Type_view",   self.views_expected, "Default_view")
        self.check_select_field(s, "Type_list",   self.lists_expected, "Default_list")
        self.check_select_field(s, "view_choice", self.views_expected, "Type_view")
        return

    # Edit/view type view
    def test_view_edit_type_view(self):
        u = collection_entity_edit_url(
            coll_id="coll1", type_id="_view", entity_id="Type_view",
            view_id="View_view", action="edit"
            )
        s = self.get_page(u)
        expect_field_choices = (
            [ "Entity_comment"
            , "Entity_id"
            , "Entity_label"
            , "Entity_type"
            , "Type_aliases"
            , "Type_comment"
            , "Type_id"
            , "Type_label"
            , "Type_list"
            , "Type_uri"
            , "Type_view"
            ])
        expect_fields = (
            [ "Type_id"
            , "Type_label"
            , "Type_comment"
            , "Type_uri"
            , "Type_view"
            , "Type_list"
            , "Type_aliases"
            ])
        self.check_view_fields(s, expect_fields, expect_field_choices)
        self.check_select_field(s, "view_choice", self.views_expected, "View_view")
        return

    # Edit/view type list
    def test_view_edit_type_list(self):
        u = collection_entity_edit_url(
            coll_id="coll1", type_id="_list", entity_id="Type_list",
            view_id="List_view", action="edit"
            )
        s = self.get_page(u)
        self.check_input_type_value(s, "entity_id", "text", "Type_list")
        self.check_input_type_value(s, "List_label", "text", None)
        self.check_input_type_value(s, "List_comment", "textarea", None)
        self.check_input_type_value(s, "List_entity_selector", "text", "'annal:Type' in [@type]")
        self.check_input_type_value(s, "List_target_type", "text", "annal:Type")
        self.check_select_field(
            s, "List_type", self.list_types_expected, "List"
            )
        self.check_select_field(
            s, "List_default_type", self.types_expected, "_type"
            )
        self.check_select_field(
            s, "List_default_view", self.views_expected, "Type_view"
            )
        return

    # --------------------------------------------------------------------------
    # Test site data for lists
    # --------------------------------------------------------------------------

    # Test type / list / view / field consistency for RecordList
    def test_recordlist_type_list_view(self):
        self.check_type_list_view("_list", "List_list", "List_view", ANNAL.CURIE.List)
        return

    # List lists using list list
    def test_list_list(self):
        u = collection_entity_list_url(
            coll_id="coll1", list_id="List_list", type_id="_list", scope='all'
            )
        s = self.get_page(u)

        self.assertEqual(s.title.string, "Collection coll1")
        self.assertEqual(s.h3.string, "List lists")
        self.check_input_type_value(s, "search_for", "text", "")
        self.check_select_field(s, "list_choice", self.lists_expected, "List_list")

        thead = s.form.find("div", class_="thead").find("div", class_="row").find_all("div", class_="columns")
        self.assertEqual(thead[0].span.string, "Id")
        self.assertEqual(thead[1].span.string, "Label")

        trows_expected = (
            [ [ "_list/_initial_values",    ["_initial_values",     None] ]
            , [ "_list/BibEntry_list",      ["BibEntry_list",       "List bibliographic entries"] ]
            , [ "_list/Default_list",       ["Default_list",        "List entities"] ]
            , [ "_list/Default_list_all",   ["Default_list_all",    "List entities with type information"] ]
            , [ "_list/Field_group_list",   ["Field_group_list",    "List field groups"] ]
            , [ "_list/Field_list",         ["Field_list",          "List fields"] ]
            , [ "_list/List_list",          ["List_list",           "List lists"] ]
            , [ "_list/Type_list",          ["Type_list",           "List types"] ]
            , [ "_list/User_list",          ["User_list",           "List users"] ]
            , [ "_list/View_list",          ["View_list",           "List views"] ]
            , [ "_list/list1",              ["list1",               "RecordList coll1/list1"] ]
            , [ "_list/list2",              ["list2",               "RecordList coll1/list2"] ]   
            ])
        self.check_list_row_data(s, trows_expected)
        return

    # Create/edit list using list view
    def test_list_edit_new(self):
        u = collection_entity_edit_url(
            coll_id="coll1", type_id="_list",
            view_id="List_view", action="new"
            )
        s = self.get_page(u)
        self.check_input_type_value(s, "entity_id", "text", None)
        self.check_input_type_value(s, "List_label", "text", None)
        self.check_input_type_value(s, "List_comment", "textarea", None)
        self.check_select_field(s, "List_default_type", self.types_expected, "Default_type")
        self.check_select_field(s, "List_default_view", self.views_expected, "Default_view")
        self.check_input_type_value(s, "List_entity_selector", "text", "ALL")
        self.check_input_type_value(s, "List_target_type", "text", "")
        self.check_select_field(s, "view_choice", self.views_expected, "List_view")
        return

    # Edit/view list view
    def test_view_edit_list_view(self):
        u = collection_entity_edit_url(
            coll_id="coll1", type_id="_view", entity_id="List_view",
            view_id="View_view", action="edit"
            )
        s = self.get_page(u)
        expect_field_choices = (
            [ "Entity_comment"
            , "Entity_id"
            , "Entity_label"
            , "Entity_type"
            , "List_comment"
            , "List_default_type"
            , "List_default_view"
            , "List_entity_selector"
            , "List_id"
            , "List_label"
            , "List_repeat_fields"
            , "List_target_type"
            , "List_type"
            ])
        expect_fields = (
            [ "List_id"
            , "List_type"
            , "List_label"
            , "List_comment"
            , "List_default_type"
            , "List_default_view"
            , "List_entity_selector"
            , "List_target_type"
            , "List_repeat_fields"
            ])
        self.check_view_fields(s, expect_fields, expect_field_choices)
        self.check_select_field(s, "view_choice", self.views_expected, "View_view")
        return

    # Edit/view list list
    def test_view_edit_list_list(self):
        u = collection_entity_edit_url(
            coll_id="coll1", type_id="_list", entity_id="List_list",
            view_id="List_view", action="edit"
            )
        s = self.get_page(u)
        self.check_input_type_value(s, "entity_id", "text", "List_list")
        self.check_input_type_value(s, "List_label", "text", None)
        self.check_input_type_value(s, "List_comment", "textarea", None)
        self.check_input_type_value(s, "List_entity_selector", "text", "'annal:List' in [@type]")
        self.check_input_type_value(s, "List_target_type", "text", "annal:List")
        self.check_select_field(
            s, "List_type", self.list_types_expected, "List"
            )
        self.check_select_field(
            s, "List_default_type", self.types_expected, "_list"
            )
        self.check_select_field(
            s, "List_default_view", self.views_expected, "List_view"
            )
        return

    # --------------------------------------------------------------------------
    # Test site data for views
    # --------------------------------------------------------------------------

    # Test type / list / view / field consistency for RecordView
    def test_recordlist_type_view_view(self):
        self.check_type_list_view("_view", "View_list", "View_view", ANNAL.CURIE.View)
        return

    # List views using view list
    def test_view_list(self):
        u = collection_entity_list_url(
            coll_id="coll1", list_id="View_list", type_id="_view", scope='all'
            )
        s = self.get_page(u)

        self.assertEqual(s.title.string, "Collection coll1")
        self.assertEqual(s.h3.string, "List views")
        self.check_input_type_value(s, "search_for", "text", "")
        self.check_select_field(s, "list_choice", self.lists_expected, "View_list")

        thead = s.form.find("div", class_="thead").find("div", class_="row").find_all("div", class_="columns")
        self.assertEqual(thead[0].span.string, "Id")
        self.assertEqual(thead[1].span.string, "Label")

        trows_expected = (
            [ [ "_view/_initial_values",    ["_initial_values",     None] ]
            , [ "_view/BibEntry_view",      ["BibEntry_view",       "Bibliographic metadata"] ]
            , [ "_view/Default_view",       ["Default_view",        "Default record view"] ]
            , [ "_view/Field_group_view",   ["Field_group_view",    "Field group view description"] ]
            , [ "_view/Field_view",         ["Field_view",          "View description for view field description"] ]
            , [ "_view/List_view",          ["List_view",           "View description for record list description"] ]
            , [ "_view/Type_view",          ["Type_view",           "View description for record type description"] ]
            , [ "_view/User_view",          ["User_view",           "User permissions view"] ]
            , [ "_view/View_view",          ["View_view",           "View description for record view description"] ]
            , [ "_view/view1",              ["view1",               "RecordView coll1/view1"] ]
            , [ "_view/view2",              ["view2",               "RecordView coll1/view2"] ]   
            ])
        self.check_list_row_data(s, trows_expected)
        return

    # Create/edit view using view view
    def test_view_edit_new(self):
        u = collection_entity_edit_url(
            coll_id="coll1", type_id="_view",
            view_id="View_view", action="new"
            )
        s = self.get_page(u)
        self.check_input_type_value(s, "entity_id", "text", None)
        self.check_input_type_value(s, "View_label", "text", None)
        self.check_input_type_value(s, "View_comment", "textarea", None)
        self.check_input_type_value(s, "View_target_type", "text", None)
        self.check_input_type_value(s, "View_edit_view", "checkbox", "Yes")
        self.check_select_field(s, "view_choice", self.views_expected, "View_view")
        return

    # Edit/view view view
    def test_view_edit_view_view(self):
        u = collection_entity_edit_url(
            coll_id="coll1", type_id="_view", entity_id="View_view",
            view_id="View_view", action="edit"
            )
        s = self.get_page(u)
        expect_field_choices = get_site_view_fields_sorted()
        expect_fields        = (
            [ "View_id"
            , "View_label"
            , "View_comment"
            , "View_target_type"
            , "View_edit_view"
            , "View_fields"
            ])
        self.check_view_fields(s, expect_fields, expect_field_choices)
        self.check_select_field(s, "view_choice", self.views_expected, "View_view")
        return

    # Edit/view view list
    def test_view_edit_view_list(self):
        u = collection_entity_edit_url(
            coll_id="coll1", type_id="_list", entity_id="View_list",
            view_id="List_view", action="edit"
            )
        s = self.get_page(u)
        self.check_input_type_value(s, "entity_id", "text", "View_list")
        self.check_input_type_value(s, "List_label", "text", None)
        self.check_input_type_value(s, "List_comment", "textarea", None)
        self.check_select_field(
            s, "List_type", self.list_types_expected, "List"
            )
        self.check_select_field(
            s, "List_default_type", self.types_expected, "_view"
            )
        self.check_select_field(
            s, "List_default_view", self.views_expected, "View_view"
            )
        self.check_input_type_value(
            s, "List_entity_selector", "text", "'annal:View' in [@type]"
            )
        self.check_input_type_value(
            s, "List_target_type", "text", "annal:View"
            )
        return

    # --------------------------------------------------------------------------
    # Test site data for field groups
    # --------------------------------------------------------------------------

    # Test type / list / view / field consistency for RecordGroup
    def test_recordlist_type_group_view(self):
        self.check_type_list_view("_group", "Field_group_list", "Field_group_view", ANNAL.CURIE.Field_group)
        return

    # List groups using group list
    def test_group_list(self):
        u = collection_entity_list_url(
            coll_id="coll1", list_id="Field_group_list", type_id="_group", scope="all"
            )
        s = self.get_page(u)

        self.assertEqual(s.title.string, "Collection coll1")
        self.assertEqual(s.h3.string, "List field groups")
        self.check_input_type_value(s, "search_for", "text", "")
        self.check_select_field(s, "list_choice", self.lists_expected, "Field_group_list")

        thead = s.form.find("div", class_="thead").find("div", class_="row").find_all("div", class_="columns")
        self.assertEqual(thead[0].span.string, "Id")
        self.assertEqual(thead[1].span.string, "Label")

        trows_expected = (
            [ [ "_group/Bib_book_group",        ["Bib_book_group",        "BibEntry book fields"] ]
            , [ "_group/Bib_identifier_group",  ["Bib_identifier_group",  "BibEntry identifier fields"] ]
            , [ "_group/Bib_journal_group",     ["Bib_journal_group",     "BibEntry journal fields"] ]
            , [ "_group/Bib_license_group",     ["Bib_license_group",     "BibEntry license fields"] ]
            , [ "_group/Bib_person_group",      ["Bib_person_group",      "BibEntry person fields"] ]
            , [ "_group/Bib_publication_group", ["Bib_publication_group", "BibEntry publication detail fields"] ]
            , [ "_group/Group_field_group",     ["Group_field_group",     "Group field description"] ]
            , [ "_group/List_field_group",      ["List_field_group",      "List field description"] ]
            , [ "_group/Type_alias_group",      ["Type_alias_group",      "Field alias"] ]
            , [ "_group/View_field_group",      ["View_field_group",      "View field description"] ]
            ])
        self.check_list_row_data(s, trows_expected)
        return

    # Create/edit group using group view
    def test_group_edit_new(self):
        u = collection_entity_edit_url(
            coll_id="coll1", type_id="_group",
            view_id="Field_group_view", action="new"
            )
        s = self.get_page(u)
        self.check_input_type_value(s, "entity_id", "text", None)
        self.check_input_type_value(s, "Group_label", "text", None)
        self.check_input_type_value(s, "Group_comment", "textarea", None)
        self.check_input_type_value(s, "Group_target_type", "text", None)
        self.check_select_field(
            s, "view_choice", self.views_expected, "Field_group_view"
            )
        return

    # Edit/view group view
    def test_view_edit_group_view(self):
        u = collection_entity_edit_url(
            coll_id="coll1", type_id="_view", entity_id="Field_group_view",
            view_id="View_view", action="edit"
            )
        s = self.get_page(u)
        expect_field_choices = (
            [ "Entity_comment"
            , "Entity_id"
            , "Entity_label"
            , "Entity_type"
            , "Group_comment"
            , "Group_field_placement"
            , "Group_field_property"
            , "Group_field_sel"
            , "Group_id"
            , "Group_label"
            , "Group_repeat_fields"
            , "Group_target_type"
            , "Type_alias_source"
            , "Type_alias_target"
            ])
        expect_fields = (
            [ "Group_id"
            , "Group_label"
            , "Group_comment"
            , "Group_target_type"
            , "Group_repeat_fields"
            ])
        self.check_view_fields(s, expect_fields, expect_field_choices)
        self.check_select_field(s, "view_choice", self.views_expected, "View_view")
        return

    # Edit/view group list
    def test_view_edit_group_list(self):
        u = collection_entity_edit_url(
            coll_id="coll1", type_id="_list", entity_id="Field_group_list",
            view_id="List_view", action="edit"
            )
        s = self.get_page(u)
        self.check_input_type_value(s, "entity_id", "text", "Field_group_list")
        self.check_input_type_value(s, "List_label", "text", None)
        self.check_input_type_value(s, "List_comment", "textarea", None)
        self.check_input_type_value(
            s, "List_entity_selector", "text", "'annal:Field_group' in [@type]"
            )
        self.check_input_type_value(
            s, "List_target_type", "text", "annal:Field_group"
            )
        self.check_select_field(
            s, "List_type", self.list_types_expected, "List"
            )
        self.check_select_field(
            s, "List_default_type", self.types_expected, "_group"
            )
        self.check_select_field(
            s, "List_default_view", self.views_expected, "Field_group_view"
            )
        return

    # --------------------------------------------------------------------------
    # Test site data for record fields
    # --------------------------------------------------------------------------

    # Test type / list / view / field consistency for RecordField
    def test_recordlist_type_field_view(self):
        self.check_type_list_view("_field", "Field_list", "Field_view", ANNAL.CURIE.Field)
        return

    # List fields using field list
    def test_field_list(self):
        u = collection_entity_list_url(
            coll_id="coll1", list_id="Field_list", type_id="_field", scope="all"
            )
        s = self.get_page(u)

        self.assertEqual(s.title.string, "Collection coll1")
        self.assertEqual(s.h3.string, "List fields")
        self.check_input_type_value(s, "search_for", "text", "")
        self.check_select_field(s, "list_choice", self.lists_expected, "Field_list")

        thead = (
            s.form.find("div", class_="thead")
            .find("div", class_="row")
            .find_all("div", class_="columns")
            )
        self.assertEqual(thead[0].span.string, "Id")
        self.assertEqual(thead[1].span.string, "Field value type")
        self.assertEqual(thead[2].span.string, "Label")

        trows_expected = (
            [ [ "_field/_initial_values",           ["_initial_values",   "annal:Text", None     ] ]
            , [ "_field/Bib_address",               ["Bib_address",       "annal:Text", "Address"] ]
            , [ "_field/Bib_alternate",             ["Bib_alternate",     "annal:Text"    ] ]
            , [ "_field/Bib_authors",               ["Bib_authors",       "bib:Authors"   ] ]
            , [ "_field/Bib_bookentry",             ["Bib_bookentry",     "bib:BookEntry" ] ]
            , [ "_field/Bib_booktitle",             ["Bib_booktitle",     "annal:Text"    ] ]
            , [ "_field/Bib_chapter",               ["Bib_chapter",       "annal:Text"    ] ]
            , [ "_field/Bib_description",           ["Bib_description",   "annal:Text"    ] ]
            , [ "_field/Bib_edition",               ["Bib_edition",       "annal:Text"    ] ]
            , [ "_field/Bib_editors",               ["Bib_editors",       "bib:Editors"   ] ]
            , [ "_field/Bib_eprint",                ["Bib_eprint",        "annal:Text"    ] ]
            , [ "_field/Bib_firstname",             ["Bib_firstname",     "annal:Text"    ] ]
            , [ "_field/Bib_howpublished",          ["Bib_howpublished",  "annal:Text"    ] ]
            , [ "_field/Bib_id",                    ["Bib_id",            "annal:Text"    ] ]
            , [ "_field/Bib_idanchor",              ["Bib_idanchor",      "annal:Text"    ] ]
            , [ "_field/Bib_identifiers",           ["Bib_identifiers",   "bib:Identifiers" ] ]
            , [ "_field/Bib_idtype",                ["Bib_idtype",        "annal:Text"    ] ]
            , [ "_field/Bib_institution",           ["Bib_institution",   "annal:Text"    ] ]
            , [ "_field/Bib_journal",               ["Bib_journal",       "bib:Journal"   ] ]
            , [ "_field/Bib_jurisdiction",          ["Bib_jurisdiction",  "annal:Text"    ] ]
            , [ "_field/Bib_lastname",              ["Bib_lastname",      "annal:Text"    ] ]
            , [ "_field/Bib_license",               ["Bib_license",       "bib:Licenses"  ] ]
            , [ "_field/Bib_month",                 ["Bib_month",         "annal:Text"    ] ]
            , [ "_field/Bib_name",                  ["Bib_name",          "annal:Text"    ] ]
            , [ "_field/Bib_note",                  ["Bib_note",          "annal:Longtext" ] ]
            , [ "_field/Bib_number",                ["Bib_number",        "annal:Text"    ] ]
            , [ "_field/Bib_organization",          ["Bib_organization",  "annal:Text"    ] ]
            , [ "_field/Bib_pages",                 ["Bib_pages",         "annal:Text"    ] ]
            , [ "_field/Bib_publication_details",   ["Bib_publication_details"            ] ]
            , [ "_field/Bib_publisher",             ["Bib_publisher",     "annal:Text"    ] ]
            , [ "_field/Bib_school",                ["Bib_school",        "annal:Text"    ] ]
            , [ "_field/Bib_shortcode",             ["Bib_shortcode",     "annal:Text"    ] ]
            , [ "_field/Bib_title",                 ["Bib_title",         "annal:Text"    ] ]
            , [ "_field/Bib_type",                  ["Bib_type",          "annal:Slug"    ] ]
            , [ "_field/Bib_url",                   ["Bib_url",           "annal:Text"    ] ]
            , [ "_field/Bib_volume",                ["Bib_volume",        "annal:Text"    ] ]
            , [ "_field/Bib_year",                  ["Bib_year",          "annal:Text"    ] ]
            , [ "_field/Entity_comment",            ["Entity_comment"                     ] ]
            , [ "_field/Entity_id",                 ["Entity_id"                          ] ]
            , [ "_field/Entity_label",              ["Entity_label"                       ] ]
            , [ "_field/Entity_type",               ["Entity_type"                        ] ]
            , [ "_field/Field_comment",             ["Field_comment"                      ] ]
            , [ "_field/Field_default",             ["Field_default"                      ] ]
            , [ "_field/Field_entity_type",         ["Field_entity_type"                  ] ]
            , [ "_field/Field_groupref",            ["Field_groupref"                     ] ]
            , [ "_field/Field_id",                  ["Field_id"                           ] ]
            , [ "_field/Field_label",               ["Field_label"                        ] ]
            , [ "_field/Field_missing",             ["Field_missing"                      ] ]
            , [ "_field/Field_placeholder",         ["Field_placeholder"                  ] ]
            , [ "_field/Field_placement",           ["Field_placement"                    ] ]
            , [ "_field/Field_property",            ["Field_property"                     ] ]
            , [ "_field/Field_render",              ["Field_render"                       ] ]
            , [ "_field/Field_repeat_label_add",    ["Field_repeat_label_add"             ] ]
            , [ "_field/Field_repeat_label_delete", ["Field_repeat_label_delete"          ] ]
            , [ "_field/Field_restrict",            ["Field_restrict"                     ] ]
            , [ "_field/Field_type",                ["Field_type"                         ] ]
            , [ "_field/Field_typeref",             ["Field_typeref"                      ] ]
            , [ "_field/Group_comment",             ["Group_comment"                      ] ]
            , [ "_field/Group_field_placement",     ["Group_field_placement"              ] ]
            , [ "_field/Group_field_property",      ["Group_field_property"               ] ]
            , [ "_field/Group_field_sel",           ["Group_field_sel"                    ] ]
            , [ "_field/Group_id",                  ["Group_id"                           ] ]
            , [ "_field/Group_label",               ["Group_label"                        ] ]
            , [ "_field/Group_repeat_fields",       ["Group_repeat_fields"                ] ]
            , [ "_field/Group_target_type",         ["Group_target_type"                  ] ]
            , [ "_field/List_choice",               ["List_choice"                        ] ]
            , [ "_field/List_comment",              ["List_comment"                       ] ]
            , [ "_field/List_default_type",         ["List_default_type"                  ] ]
            , [ "_field/List_default_view",         ["List_default_view"                  ] ]
            , [ "_field/List_entity_selector",      ["List_entity_selector"               ] ]
            , [ "_field/List_id",                   ["List_id"                            ] ]
            , [ "_field/List_label",                ["List_label"                         ] ]
            , [ "_field/List_repeat_fields",        ["List_repeat_fields"                 ] ]
            , [ "_field/List_target_type",          ["List_target_type"                   ] ]
            , [ "_field/List_type",                 ["List_type"                          ] ]
            , [ "_field/Type_alias_source",         ["Type_alias_source"                  ] ]
            , [ "_field/Type_alias_target",         ["Type_alias_target"                  ] ]
            , [ "_field/Type_aliases",              ["Type_aliases"                       ] ]
            , [ "_field/Type_comment",              ["Type_comment"                       ] ]
            , [ "_field/Type_id",                   ["Type_id"                            ] ]
            , [ "_field/Type_label",                ["Type_label"                         ] ]
            , [ "_field/Type_list",                 ["Type_list"                          ] ]
            , [ "_field/Type_uri",                  ["Type_uri"                           ] ]
            , [ "_field/Type_view",                 ["Type_view"                          ] ]
            , [ "_field/User_description",          ["User_description"                   ] ]
            , [ "_field/User_id",                   ["User_id"                            ] ]
            , [ "_field/User_name",                 ["User_name"                          ] ]
            , [ "_field/User_permissions",          ["User_permissions"                   ] ]
            , [ "_field/User_uri",                  ["User_uri"                           ] ]
            , [ "_field/View_choice",               ["View_choice"                        ] ]
            , [ "_field/View_comment",              ["View_comment"                       ] ]
            , [ "_field/View_edit_view",            ["View_edit_view"                     ] ]
            , [ "_field/View_fields",               ["View_fields"                        ] ]
            , [ "_field/View_id",                   ["View_id"                            ] ]
            , [ "_field/View_label",                ["View_label"                         ] ]
            , [ "_field/View_target_type",          ["View_target_type"                   ] ]
            ])
        self.check_list_row_data(s, trows_expected)

    # Create/edit field using field view
    def test_field_edit_new(self):
        u = collection_entity_edit_url(
            coll_id="coll1", type_id="_field",
            view_id="Field_view", action="new"
            )
        s = self.get_page(u)
        self.check_input_type_value(s, "entity_id", "text", None)
        self.check_input_type_value(s, "Field_type", "text", "annal:Text")
        self.check_input_type_value(s, "Field_label", "text", "")
        self.check_input_type_value(s, "Field_comment", "textarea", None)
        self.check_input_type_value(s, "Field_placeholder", "text", "")
        self.check_input_type_value(s, "Field_property", "text", "")
        # self.check_input_type_value(s, "Field_placement", "text", "")
        self.check_select_field(
            s, "Field_placement", 
            ["(field position and size)"]+self.placements_expected, 
            "(field position and size)"
            )
        self.check_input_type_value(s, "Field_default", "text", None)
        self.check_input_type_value(s, "Field_entity_type", "text", "")
        self.check_input_type_value(s, "Field_restrict", "text", "")
        self.check_select_field(
            s, "Field_render",   self.render_types_expected, "Text"
            )
        self.check_select_field(
            s, "Field_typeref", 
            ["(no type selected)"]+self.types_expected, 
            "(no type selected)"
            )
        self.check_select_field(
            s, "Field_groupref",
            ["(no field group selected)"]+self.grouprefs_expected, 
            "(no field group selected)"
            )
        self.check_select_field(
            s, "view_choice", self.views_expected, "Field_view"
            )
        return

    # Edit/view field view
    def test_view_edit_field_view(self):
        u = collection_entity_edit_url(
            coll_id="coll1", type_id="_view", entity_id="Field_view",
            view_id="View_view", action="edit"
            )
        s = self.get_page(u)
        expect_field_choices = (
            [ "Entity_comment"
            , "Entity_id"
            , "Entity_label"
            , "Entity_type"
            , "Field_comment"
            , "Field_default"
            , "Field_entity_type"
            , "Field_groupref"
            , "Field_id"
            , "Field_label"
            , "Field_missing"
            , "Field_placeholder"
            , "Field_placement"
            , "Field_property"
            , "Field_render"
            , "Field_repeat_label_add"
            , "Field_repeat_label_delete"
            , "Field_restrict"
            , "Field_type"
            , "Field_typeref"
            ])
        expect_fields = (
            [ "Field_id"
            , "Field_type"
            , "Field_render"
            , "Field_placement"
            , "Field_label"
            , "Field_comment"
            , "Field_placeholder"
            , "Field_property"
            , "Field_default"
            , "Field_entity_type"
            , "Field_typeref"
            , "Field_restrict"
            , "Field_groupref"
            , "Field_repeat_label_add"
            , "Field_repeat_label_delete"
            ])
        self.check_view_fields(s, expect_fields, expect_field_choices)
        self.check_select_field(s, "view_choice", self.views_expected, "View_view")
        return

    # Edit/view field list
    def test_view_edit_field_list(self):
        u = collection_entity_edit_url(
            coll_id="coll1", type_id="_list", entity_id="Field_list",
            view_id="List_view", action="edit"
            )
        s = self.get_page(u)
        self.check_input_type_value(s, "entity_id", "text", "Field_list")
        self.check_input_type_value(s, "List_label", "text", None)
        self.check_input_type_value(s, "List_comment", "textarea", None)
        self.check_input_type_value(s, "List_target_type", "text", "annal:Field")
        self.check_input_type_value(
            s, "List_entity_selector", "text", "'annal:Field' in [@type]"
            )
        self.check_select_field(
            s, "List_default_type", self.types_expected, "_field"
            )
        self.check_select_field(
            s, "List_default_view", self.views_expected, "Field_view"
            )
        return

    # --------------------------------------------------------------------------
    # Test site data for user permissions
    # --------------------------------------------------------------------------

    # Test type / list / view / field consistency for user
    def test_recorduser_type_field_view(self):
        self.check_type_list_view("_user", "User_list", "User_view", ANNAL.CURIE.User)
        return

    # List users using user list
    def test_user_list(self):
        u = collection_entity_list_url(
            coll_id="coll1", list_id="User_list", type_id="_user", scope="all"
            )
        s = self.get_page(u)

        self.assertEqual(s.title.string, "Collection coll1")
        self.assertEqual(s.h3.string, "List users")
        self.check_input_type_value(s, "search_for", "text", "")
        self.check_select_field(s, "list_choice", self.lists_expected, "User_list")

        thead = s.form.find("div", class_="thead").find("div", class_="row").find_all("div", class_="columns")
        self.assertEqual(thead[0].span.string, "User Id")
        self.assertEqual(thead[1].span.string, "URI")
        self.assertEqual(thead[2].span.string, "Permissions")

        trows_expected = (
            [ [ "_user/_default_user_perms",  
                [ "_default_user_perms",  "annal:User/_default_user_perms"
                , "VIEW"]
              ]
            , [ "_user/_unknown_user_perms",
                [ "_unknown_user_perms",  "annal:User/_unknown_user_perms"
                , "VIEW"
                ] 
              ]
            , [ "_user/admin", 
                [ "admin", "mailto:admin@localhost"
                , "VIEW CREATE UPDATE DELETE CONFIG CREATE_COLL DELETE_COLL ADMIN"
                ]
              ]
            , [ "_user/testuser",
                [ "testuser", "mailto:testuser@test.example.com"
                , "VIEW CREATE UPDATE DELETE CONFIG ADMIN"
                ]
              ]
            ])
        self.check_list_row_data(s, trows_expected)

    # Create/edit user using user view
    def test_user_edit_new(self):
        u = collection_entity_edit_url(
            coll_id="coll1", type_id="_user",
            view_id="User_view", action="new"
            )
        s = self.get_page(u)
        self.check_input_type_value(s, "entity_id", "text", None)
        self.check_input_type_value(s, "User_name", "text", None)
        self.check_input_type_value(s, "User_description", "textarea", None)
        self.check_input_type_value(s, "User_uri", "text", None)
        self.check_input_type_value(s, "User_permissions", "text", None)
        self.check_select_field(s, "view_choice", self.views_expected, "User_view")
        return

    # Edit/view user view
    def test_view_edit_user_view(self):
        u = collection_entity_edit_url(
            coll_id="coll1", type_id="_view", entity_id="User_view",
            view_id="View_view", action="edit"
            )
        s = self.get_page(u)
        expect_field_choices = (
            [ "Entity_comment"
            , "Entity_id"
            , "Entity_label"
            , "Entity_type"
            , "User_description"
            , "User_id"
            , "User_name"
            , "User_permissions"
            , "User_uri"
            ])
        expect_fields = (
            [ "User_id"
            , "User_name"
            , "User_description"
            , "User_uri"
            , "User_permissions"
            ])
        self.check_view_fields(s, expect_fields, expect_field_choices)
        self.check_select_field(s, "view_choice", self.views_expected, "View_view")
        return

    # Edit/view user list
    def test_view_edit_user_list(self):
        u = collection_entity_edit_url(
            coll_id="coll1", type_id="_list", entity_id="User_list",
            view_id="List_view", action="edit"
            )
        s = self.get_page(u)
        self.check_input_type_value(s, "entity_id", "text", "User_list")
        self.check_input_type_value(s, "List_label", "text", None)
        self.check_input_type_value(s, "List_comment", "textarea", None)
        self.check_input_type_value(s, "List_target_type", "text", "annal:User")
        self.check_input_type_value(
            s, "List_entity_selector", "text", "'annal:User' in [@type]"
            )
        self.check_select_field(
            s, "List_default_type", self.types_expected, "_user"
            )
        self.check_select_field(
            s, "List_default_view", self.views_expected, "User_view"
            )
        return

# End.



"""
This module defines a class that is used to gather information about an entity
list or view display that is needed to process various kinds of HTTP request.

The intent of this module is to collect and isolate various housekeeping functions
into a common module to avoid repetition of logic and reduce code clutter in
the various Annalist view processing handlers.
"""

__author__      = "Graham Klyne (GK@ACM.ORG)"
__copyright__   = "Copyright 2014, G. Klyne"
__license__     = "MIT (http://opensource.org/licenses/MIT)"

import logging
log = logging.getLogger(__name__)

from django.conf                    import settings
from django.http                    import HttpResponse
from django.http                    import HttpResponseRedirect
from django.core.urlresolvers       import resolve, reverse

from annalist.identifiers           import RDF, RDFS, ANNAL
from annalist                       import message

from annalist.models.entitytypeinfo import EntityTypeInfo
from annalist.models.collection     import Collection
from annalist.models.recordtype     import RecordType
from annalist.models.recordtypedata import RecordTypeData
from annalist.models.recordlist     import RecordList
from annalist.models.recordview     import RecordView

from annalist.views.uri_builder     import uri_with_params

#   -------------------------------------------------------------------------------------------
#
#   Table of authorization symbols added to the display context according to 
#   permissions help by the requesting user
#
#   -------------------------------------------------------------------------------------------

authorization_map = (
    { "auth_create":        ["CREATE"]
    , "auth_delete":        ["DELETE"]
    , "auth_update":        ["UPDATE"]
    , "auth_config":        ["CONFIG"]
    , "auth_create_coll":   ["CREATE_COLLECTION", "ADMIN"]
    , "auth_delete_coll":   ["DELETE_COLLECTION", "ADMIN"]
    })

#   -------------------------------------------------------------------------------------------
#
#   Display information class
#
#   -------------------------------------------------------------------------------------------

class DisplayInfo(object):
    """
    This class collects and organizes common information needed to process various
    kinds of view requests.

    A number of methods are provided that collect different kinds of information,
    allowing the calling method flexibility over what information is actually 
    gathered.  All methods follow a common pattern loosely modeled on an error
    monad, which uses a Django HttpResponse object to record the first problem
    found in the information gathering chain.  Once an error has been detected, 
    subsequent methods do not update the display information, but simply return
    the error response object.

    The information gathering methods do have some dependencies and must be
    invoked in a sequence that ensures the dependencies are satisfied.
    """

    def __init__(self, view, action):
        self.view           = view
        self.action         = action
        # Default no permissions:
        self.authorizations = dict([(k, False) for k in authorization_map])
        self.reqhost        = None
        self.site           = None
        self.sitedata       = None
        self.coll_id        = None
        self.collection     = None
        self.type_id        = None
        self.entitytypeinfo = None
        self.list_id        = None
        self.recordlist     = None
        self.view_id        = None
        self.recordview     = None
        self.entity_id      = None
        # self.entitydata     = None
        self.http_response  = None
        return

    def get_site_info(self, reqhost):
        if not self.http_response:
            self.reqhost        = reqhost
            self.site           = self.view.site(host=reqhost)
            self.sitedata       = self.view.site_data()
        return self.http_response

    def get_coll_info(self, coll_id):
        """
        Check collection identifier, and get reference to collection object.
        """
        assert (self.site is not None)
        if not self.http_response:
            if not Collection.exists(self.site, coll_id):
                self.http_response = self.view.error(
                    dict(self.view.error404values(),
                        message=message.COLLECTION_NOT_EXISTS%{'id': coll_id}
                        )
                    )
            else:
                self.coll_id    = coll_id
                self.collection = Collection.load(self.site, coll_id)
        return self.http_response

    def get_type_info(self, type_id):
        """
        Check type identifier, and get reference to type information object.
        """
        if not self.http_response:
            assert ((self.site and self.collection) is not None)
            if type_id:
                self.type_id        = type_id
                self.entitytypeinfo = EntityTypeInfo(self.site, self.collection, type_id)
                if not self.entitytypeinfo.recordtype:
                    # log.warning("DisplayInfo.get_type_data: RecordType %s not found"%type_id)
                    self.http_response = self.view.error(
                        dict(self.view.error404values(),
                            message=message.RECORD_TYPE_NOT_EXISTS%(
                                {'id': type_id, 'coll_id': self.coll_id})
                            )
                        )
        return self.http_response

    def get_list_info(self, list_id):
        """
        Retrieve list definition to use for display
        """
        if not self.http_response:
            assert ((self.site and self.collection) is not None)
            assert list_id
            if not RecordList.exists(self.collection, list_id, self.site):
                log.warning("DisplayInfo.get_list_info: RecordList %s not found"%list_id)
                self.http_response = self.view.error(
                    dict(self.view.error404values(),
                        message=message.RECORD_LIST_NOT_EXISTS%(
                            {'id': list_id, 'coll_id': self.coll_id})
                        )
                    )
            else:
                self.list_id    = list_id
                self.recordlist = RecordList.load(self.collection, list_id, self.site)
                log.debug("DisplayInfo.get_list_info: %r"%(self.recordlist.get_values()))
        return self.http_response

    def get_view_info(self, view_id):
        """
        Retrieve view definition to use for display
        """
        if not self.http_response:
            assert ((self.site and self.collection) is not None)
            if not RecordView.exists(self.collection, view_id, self.site):
                log.warning("DisplayInfo.get_view_info: RecordView %s not found"%view_id)
                self.http_response = self.view.error(
                    dict(self.view.error404values(),
                        message=message.RECORD_VIEW_NOT_EXISTS%(
                            {'id': view_id, 'coll_id': self.coll_id})
                        )
                    )
            else:
                self.view_id    = view_id
                self.recordview = RecordView.load(self.collection, view_id, self.site)
                log.debug("DisplayInfo.get_view_info: %r"%(self.recordview.get_values()))
        return self.http_response

    def get_entity_info(self, action, entity_id):
        """
        Retrieve entity data to use for display
        """
        if not self.http_response:
            assert self.entitytypeinfo is not None
            if (not entity_id) and (action == "new"):
                entity_id = self.entitytypeinfo.entityclass.allocate_new_id(
                    self.entitytypeinfo.entityparent
                    )
            self.entity_id = entity_id
        return self.http_response

    def get_entity_data_not_used(self):
        """
        Retrieve entity data to use for display
        """
        if not self.http_response:
            assert self.entity_id is not None
            self.entitydata = self.entitytypeinfo.entityclass.load(
                self.entitytypeinfo.entityparent, 
                self.entity_id, 
                self.entitytypeinfo.entityaltparent)
            if self.entitydata:
                log.debug("DisplayInfo.get_entity_data: %r"%(self.entitydata.get_values()))
        return self.http_response

    def check_authorization(self, action):
        """
        If no error so far, check authorization.  Return None if all is OK,
        or HttpResonse object.

        Also, save copy of key authorizations for later rendering.
        """
        action = action or "view"
        if self.entitytypeinfo:
            permissions_map = self.entitytypeinfo.permissions_map
        else:
            permissions_map = {}
        self.http_response = (
            self.http_response or 
            self.view.form_action_auth(action, self.collection, permissions_map)
            )
        for k in authorization_map:
            for p in authorization_map[k]:
                self.authorizations[k] = (
                    self.authorizations[k] or 
                    self.view.authorize(p, self.collection) is None
                    )
        return self.http_response

    # Additonal support functions for list views

    def get_type_list_id(self, type_id):
        """
        Return default list_id for listing defined type, or None
        """
        list_id = None
        if type_id:
            if self.entitytypeinfo.recordtype:
                list_id = self.entitytypeinfo.recordtype.get(ANNAL.CURIE.type_list, None)
            else:
                log.warning("DisplayInfo.get_type_list_id no type data for %s"%(type_id))
        return list_id

    def get_list_id(self, type_id, list_id):
        """
        Return supplied list_id if defined, otherwise find default list_id for
        entity type or collection (unless an error has been detected).
        """
        if not self.http_response:
            list_id = (
                list_id or 
                self.get_type_list_id(type_id) or
                self.collection.get_default_list() or
                ("Default_list" if type_id else "Default_list_all")
                )
            if not list_id:
                log.warning("get_list_id: %s, type_id %s"%(list_id, type_id))
        return list_id

    def get_list_view_id(self):
        return self.recordlist.get(ANNAL.CURIE.default_view, None) or "Default_view"

    def get_list_type_id(self):
        return self.recordlist.get(ANNAL.CURIE.default_type, None) or "Default_type"

    def check_collection_entity(self, entity_id, entity_type, msg, continuation_url={}):
        """
        Test a supplied entity_id is defined in the current collection,
        returning a URI to display a supplied error message if the test fails.

        NOTE: this function works with the generic base template base_generic.html, which
        is assumed to provide an underlay for the currently viewed page.

        entity_id           entity id that is required to be defined in the current collection.
        entity_type         specified type for entity to delete.
        msg                 message to display if the test fails.
        continuation_url    URI of page to display when the redisplayed form is closed.

        returns a URI string for use with HttpResponseRedirect to redisplay the 
        current page with the supplied message, or None if entity id is OK.
        """
        # log.info("check_collection_entity: entity_id: %s"%(entity_id))
        # log.info("check_collection_entity: entityparent: %s"%(self.entityparent.get_id()))
        # log.info("check_collection_entity: entityclass: %s"%(self.entityclass))
        redirect_uri = None
        typeinfo     = (
            self.entitytypeinfo or 
            EntityTypeInfo(self.site, self.collection, entity_type)
            )
        if not typeinfo.entityclass.exists(typeinfo.entityparent, entity_id):
            redirect_uri = (
                uri_with_params(
                    self.view.get_request_path(),
                    self.view.error_params(msg),
                    continuation_url
                    )
                )
        return redirect_uri

    def get_new_view_uri(self, coll_id, type_id):
        """
        Get URI for entity new view from list display
        """
        return self.view.view_uri(
            "AnnalistEntityNewView", 
            coll_id=coll_id, 
            view_id=self.get_list_view_id(), 
            type_id=type_id,
            action="new"
            )

    def get_edit_view_uri(self, coll_id, type_id, entity_id, action):
        """
        Get URI for entity edit or copy view from list display
        """
        return self.view.view_uri(
                "AnnalistEntityEditView", 
                coll_id=coll_id, 
                view_id=self.get_list_view_id(), 
                type_id=type_id,
                entity_id=entity_id,
                action=action
                )

    # Additonal support functions

    #@@
    # def get_type_view_id(self, type_id):
    #     """
    #     Get default view id for given type
    #     """
    #     # @@TODO: remove this
    #     view_id = None
    #     if self.type_id:
    #         if self.entitytypeinfo.recordtype:
    #             view_id  = self.entitytypeinfo.recordtype.get(ANNAL.CURIE.type_view, None)
    #         else:
    #             log.warning("DisplayInfo.get_type_view_id: no type data for %s"%(self.type_id))
    #     return view_id
    #@@

    def get_view_id(self, type_id, view_id):
        """
        Get view id or suitable default using type if defined.
        """
        if not self.http_response:
            view_id = (
                view_id or 
                self.entitytypeinfo.get_default_view_id()
                # self.get_type_view_id(type_id) or
                # self.collection.get_default_view() or
                # "Default_view"
                )
            if not view_id:
                log.warning("get_view_id: %s, type_id %s"%(view_id, self.type_id))
        return view_id

    def get_save_continuation_url(self, entity_type, entity_id, action):
        """
        Gets a URI that is based on that used to invoke the current view,
        but with a different action and specified entity_type/entity_id

        This is used to continue editing a new or copied entity after that 
        entity has been saved, or viewing an entity following an edit.
        """
        return self.view.view_uri(
                "AnnalistEntityEditView", 
                coll_id=self.coll_id,
                view_id=self.view_id,
                type_id=entity_type,
                entity_id=entity_id,
                action=action
                )

    def context_data(self):
        """
        Return dictionary of rendering context data available from the elements assembled.

        Values that are added here to the view context are used for view rendering, and
        are not passed to the entity value mapping process.
        """
        context = (
            { 'site_title':     self.sitedata["title"]
            , 'title':          self.sitedata["title"]
            , 'action':         self.action
            , 'coll_id':        self.coll_id
            , 'type_id':        self.type_id
            , 'view_id':        self.view_id
            , 'list_id':        self.list_id
            })
        context.update(self.authorizations)
        if hasattr(self.view, 'help'):
            context.update(
                { 'help_filename':  self.view.help
                })
        if self.collection:
            context.update(
                { 'title':      self.collection[RDFS.CURIE.label]
                , 'coll_label': self.collection[RDFS.CURIE.label]
                })
        if self.recordview:
            context.update(
                { 'view_label': self.recordview[RDFS.CURIE.label]
                })
        if self.recordlist:
            context.update(
                { 'list_label': self.recordlist[RDFS.CURIE.label]
                })
        return context

    def __str__(self):
        attrs = (
            [ "view"
            , "action"
            , "authorizations"
            , "reqhost"
            , "site"
            , "sitedata"
            , "coll_id"
            , "collection"
            , "type_id"
            , "entitytypeinfo"
            , "list_id"
            , "recordlist"
            , "view_id"
            , "recordview"
            , "entity_id"
            ])
        fields = ["DisplayInfo("]
        for attr in attrs:
            val = getattr(self, attr, None)
            if val is not None:
                fields.append("%s: %r"%(attr, val))
        fields.append(")")
        return ("\n  ".join(fields))

    def __repr__(self):
        return str(self)

# End.



"""
Tests for RecordType module and view

Note: this module tests for rendering specifically for RecordType values, using
type description sitedata files, and as such duplicates some tests covered by
module test_entitygenericedit.
"""

__author__      = "Graham Klyne (GK@ACM.ORG)"
__copyright__   = "Copyright 2014, G. Klyne"
__license__     = "MIT (http://opensource.org/licenses/MIT)"

import os
import unittest

import logging
log = logging.getLogger(__name__)

from django.conf                        import settings
from django.db                          import models
from django.http                        import QueryDict
from django.contrib.auth.models         import User
from django.test                        import TestCase # cf. https://docs.djangoproject.com/en/dev/topics/testing/tools/#assertions
from django.test.client                 import Client

from annalist.identifiers               import RDF, RDFS, ANNAL
from annalist                           import layout
from annalist.models.site               import Site
from annalist.models.sitedata           import SiteData
from annalist.models.collection         import Collection
from annalist.models.entitydata         import EntityData
from annalist.models.recordtype         import RecordType
from annalist.models.recordtypedata     import RecordTypeData
from annalist.models.recordview         import RecordView
from annalist.models.recordlist         import RecordList

from annalist.views.recordtypedelete    import RecordTypeDeleteConfirmedView

from tests                              import TestHost, TestHostUri, TestBasePath, TestBaseUri, TestBaseDir
from tests                              import init_annalist_test_site
from AnnalistTestCase                   import AnnalistTestCase
from entity_testutils                   import (
    site_dir, collection_dir,
    site_view_url, collection_edit_url, 
    collection_entity_view_url,
    collection_create_values,
    create_test_user
    )
from entity_testtypedata                import (
    recordtype_dir,
    recordtype_coll_url, recordtype_site_url, recordtype_url, recordtype_edit_url,
    recordtype_value_keys, recordtype_load_keys, 
    recordtype_create_values, recordtype_values, recordtype_read_values,
    recordtype_entity_view_context_data, 
    recordtype_entity_view_form_data, recordtype_delete_confirm_form_data
    )
from entity_testentitydata              import (
    entity_url, entitydata_edit_url, entitydata_list_type_url,
    default_fields, default_label, default_comment, error_label,
    layout_classes
    )

#   -----------------------------------------------------------------------------
#
#   RecordType tests
#
#   -----------------------------------------------------------------------------

class RecordTypeTest(AnnalistTestCase):
    """
    Tests for RecordType object interface
    """

    def setUp(self):
        init_annalist_test_site()
        self.testsite = Site(TestBaseUri, TestBaseDir)
        self.sitedata = SiteData(self.testsite)
        self.testcoll = Collection(self.testsite, "testcoll")
        return

    def tearDown(self):
        return

    def test_RecordTypeTest(self):
        self.assertEqual(RecordType.__name__, "RecordType", "Check RecordType class name")
        return

    def test_recordtype_init(self):
        t = RecordType(self.testcoll, "testtype", self.testsite)
        u = recordtype_coll_url(self.testsite, coll_id="testcoll", type_id="testtype")
        self.assertEqual(t._entitytype,     ANNAL.CURIE.Type)
        self.assertEqual(t._entityfile,     layout.TYPE_META_FILE)
        self.assertEqual(t._entityref,      layout.META_TYPE_REF)
        self.assertEqual(t._entityid,       "testtype")
        self.assertEqual(t._entityurl,      u)
        self.assertEqual(t._entitydir,      recordtype_dir(type_id="testtype"))
        self.assertEqual(t._values,         None)
        return

    def test_recordtype1_data(self):
        t = RecordType(self.testcoll, "type1", self.testsite)
        self.assertEqual(t.get_id(), "type1")
        self.assertEqual(t.get_type_id(), "_type")
        self.assertIn("/c/testcoll/_annalist_collection/types/type1/", t.get_url())
        self.assertEqual(TestBaseUri + "/c/testcoll/d/_type/type1/", t.get_view_url())
        t.set_values(recordtype_create_values(type_id="type1"))
        td = t.get_values()
        self.assertEqual(set(td.keys()), set(recordtype_value_keys()))
        v = recordtype_values(type_id="type1")
        self.assertDictionaryMatch(td, v)
        return

    def test_recordtype2_data(self):
        t = RecordType(self.testcoll, "type2", self.testsite)
        self.assertEqual(t.get_id(), "type2")
        self.assertEqual(t.get_type_id(), "_type")
        self.assertIn("/c/testcoll/_annalist_collection/types/type2/", t.get_url())
        self.assertEqual(TestBaseUri + "/c/testcoll/d/_type/type2/", t.get_view_url())
        t.set_values(recordtype_create_values(type_id="type2"))
        td = t.get_values()
        self.assertEqual(set(td.keys()), set(recordtype_value_keys()))
        v = recordtype_values(type_id="type2")
        self.assertDictionaryMatch(td, v)
        return

    def test_recordtype_create_load(self):
        t  = RecordType.create(self.testcoll, "type1", recordtype_create_values(type_id="type1"))
        td = RecordType.load(self.testcoll, "type1").get_values()
        v  = recordtype_read_values(type_id="type1")
        self.assertKeysMatch(td, v)
        self.assertDictionaryMatch(td, v)
        return

    def test_recordtype_default_data(self):
        t = RecordType.load(self.testcoll, "Default_type", altparent=self.testsite)
        self.assertEqual(t.get_id(), "Default_type")
        self.assertIn("/c/testcoll/_annalist_collection/types/Default_type", t.get_url())
        self.assertEqual(t.get_type_id(), "_type")
        td = t.get_values()
        self.assertEqual(set(td.keys()), set(recordtype_load_keys(type_uri=True)))
        v = recordtype_read_values(type_id="Default_type")
        v.update(
            { 'rdfs:label':     'Default record type'
            , 'rdfs:comment':   'Default record type, applied when no type is specified when creating a record.'
            , 'annal:uri':      'annal:Default_type'
            })
        self.assertDictionaryMatch(td, v)
        return

#   -----------------------------------------------------------------------------
#
#   RecordTypeEditView tests
#
#   -----------------------------------------------------------------------------

class RecordTypeEditViewTest(AnnalistTestCase):
    """
    Tests for record type edit views
    """

    def setUp(self):
        init_annalist_test_site()
        self.testsite = Site(TestBaseUri, TestBaseDir)
        self.testcoll = Collection.create(self.testsite, "testcoll", collection_create_values("testcoll"))
        self.no_options = ['(no options)']
        self.view_options    = sorted(
            [ vid for vid in self.testcoll.child_entity_ids(RecordView, self.testsite) 
                  if vid != "_initial_values"
            ])
        self.list_options    = sorted(
            [ lid for lid in self.testcoll.child_entity_ids(RecordList, self.testsite) 
                  if lid != "_initial_values"
            ])
        # For checking Location: header values...
        self.continuation_url = TestHostUri + entitydata_list_type_url(coll_id="testcoll", type_id="_type")
        # Login and permissions
        create_test_user(self.testcoll, "testuser", "testpassword")
        self.client = Client(HTTP_HOST=TestHost)
        loggedin = self.client.login(username="testuser", password="testpassword")
        self.assertTrue(loggedin)
        return

    def tearDown(self):
        return

    #   -----------------------------------------------------------------------------
    #   Helpers
    #   -----------------------------------------------------------------------------

    def _create_record_type(self, type_id, entity_id="testentity"):
        "Helper function creates record type entry with supplied type_id"
        t = RecordType.create(self.testcoll, type_id, recordtype_create_values(type_id=type_id))
        d = RecordTypeData.create(self.testcoll, type_id, {})
        e = EntityData.create(d, entity_id, {})
        return (t, d, e)

    def _check_record_type_values(self, type_id, update="RecordType"):
        "Helper function checks content of record type entry with supplied type_id"
        self.assertTrue(RecordType.exists(self.testcoll, type_id))
        t = RecordType.load(self.testcoll, type_id)
        self.assertEqual(t.get_id(), type_id)
        self.assertEqual(t.get_view_url(), TestHostUri + recordtype_url("testcoll", type_id))
        v = recordtype_values(type_id=type_id, update=update)
        self.assertDictionaryMatch(t.get_values(), v)
        return t

    def _check_context_fields(self, response, 
            type_id="(?type_id)", 
            type_label="(?type_label)",
            type_help="(?type_help)",
            type_uri="(?type_uri)",
            type_view="Default_view",
            type_list="Default_list"
            ):
        r = response
        self.assertEqual(len(r.context['fields']), 7)
        # 1st field - Id
        type_id_help = (
            "A short identifier that distinguishes this type from all other types in the same collection."
            )
        self.assertEqual(r.context['fields'][0]['field_id'], 'Type_id')
        self.assertEqual(r.context['fields'][0]['field_name'], 'entity_id')
        self.assertEqual(r.context['fields'][0]['field_label'], 'Id')
        self.assertEqual(r.context['fields'][0]['field_help'], type_id_help)
        self.assertEqual(r.context['fields'][0]['field_placeholder'], "(type id)")
        self.assertEqual(r.context['fields'][0]['field_property_uri'], "annal:id")
        self.assertEqual(r.context['fields'][0]['field_placement'].field, "small-12 medium-6 columns")
        self.assertEqual(r.context['fields'][0]['field_value_type'], "annal:Slug")
        self.assertEqual(r.context['fields'][0]['field_value'], type_id)
        self.assertEqual(r.context['fields'][0]['options'], self.no_options)
        # 2nd field - Label
        type_label_help = (
            "Short string used to describe record type when displayed"
            )
        self.assertEqual(r.context['fields'][1]['field_id'], 'Type_label')
        self.assertEqual(r.context['fields'][1]['field_name'], 'Type_label')
        self.assertEqual(r.context['fields'][1]['field_label'], 'Label')
        self.assertEqual(r.context['fields'][1]['field_help'], type_label_help)
        self.assertEqual(r.context['fields'][1]['field_placeholder'], "(label)")
        self.assertEqual(r.context['fields'][1]['field_property_uri'], "rdfs:label")
        self.assertEqual(r.context['fields'][1]['field_placement'].field, "small-12 columns")
        self.assertEqual(r.context['fields'][1]['field_value_type'], "annal:Text")
        self.assertEqual(r.context['fields'][1]['field_value'], type_label)
        self.assertEqual(r.context['fields'][1]['options'], self.no_options)
        # 3rd field - comment
        type_comment_help = (
            "Descriptive text about a record type"
            )
        type_comment_placeholder = (
            "(type description)"
            )
        self.assertEqual(r.context['fields'][2]['field_id'], 'Type_comment')
        self.assertEqual(r.context['fields'][2]['field_name'], 'Type_comment')
        self.assertEqual(r.context['fields'][2]['field_label'], 'Comment')
        self.assertEqual(r.context['fields'][2]['field_help'], type_comment_help)
        self.assertEqual(r.context['fields'][2]['field_placeholder'], type_comment_placeholder)
        self.assertEqual(r.context['fields'][2]['field_property_uri'], "rdfs:comment")
        self.assertEqual(r.context['fields'][2]['field_placement'].field, "small-12 columns")
        self.assertEqual(r.context['fields'][2]['field_value_type'], "annal:Longtext")
        self.assertEqual(r.context['fields'][2]['field_value'], type_help)
        self.assertEqual(r.context['fields'][2]['options'], self.no_options)
        # 4th field - URI
        type_uri_help = (
            "Entity type URI"
            )
        type_uri_placeholder = (
            "(URI)"
            )
        self.assertEqual(r.context['fields'][3]['field_id'], 'Type_uri')
        self.assertEqual(r.context['fields'][3]['field_name'], 'Type_uri')
        self.assertEqual(r.context['fields'][3]['field_label'], 'URI')
        self.assertEqual(r.context['fields'][3]['field_help'], type_uri_help)
        self.assertEqual(r.context['fields'][3]['field_placeholder'], type_uri_placeholder)
        self.assertEqual(r.context['fields'][3]['field_property_uri'], "annal:uri")
        self.assertEqual(r.context['fields'][3]['field_placement'].field, "small-12 columns")
        self.assertEqual(r.context['fields'][3]['field_value_type'], "annal:Identifier")
        self.assertEqual(r.context['fields'][3]['field_value'], type_uri)
        self.assertEqual(r.context['fields'][3]['options'], self.no_options)
        # 5th field - view id
        type_view_id_help = (
            "Default view id for this type"
            )
        type_view_id_placeholder = (
            "(view id)"
            )
        self.assertEqual(r.context['fields'][4]['field_id'], 'Type_view')
        self.assertEqual(r.context['fields'][4]['field_name'], 'Type_view')
        self.assertEqual(r.context['fields'][4]['field_label'], 'Default view')
        self.assertEqual(r.context['fields'][4]['field_help'], type_view_id_help)
        self.assertEqual(r.context['fields'][4]['field_placeholder'], type_view_id_placeholder)
        self.assertEqual(r.context['fields'][4]['field_property_uri'], "annal:type_view")
        self.assertEqual(r.context['fields'][4]['field_placement'].field, "small-12 medium-6 columns")
        self.assertEqual(r.context['fields'][4]['field_value_type'], "annal:View")
        self.assertEqual(r.context['fields'][4]['field_value'], type_view)
        self.assertEqual(r.context['fields'][4]['options'], self.view_options)
        # 6th field - list id
        type_list_id_help = (
            "Default list id for this type"
            )
        type_list_id_placeholder = (
            "(list id)"
            )
        self.assertEqual(r.context['fields'][5]['field_id'], 'Type_list')
        self.assertEqual(r.context['fields'][5]['field_name'], 'Type_list')
        self.assertEqual(r.context['fields'][5]['field_label'], 'Default list')
        self.assertEqual(r.context['fields'][5]['field_help'], type_list_id_help)
        self.assertEqual(r.context['fields'][5]['field_placeholder'], type_list_id_placeholder)
        self.assertEqual(r.context['fields'][5]['field_property_uri'], "annal:type_list")
        self.assertEqual(r.context['fields'][5]['field_placement'].field, "small-12 medium-6 columns")
        self.assertEqual(r.context['fields'][5]['field_value_type'], "annal:List")
        self.assertEqual(r.context['fields'][5]['field_value'], type_list)
        self.assertEqual(r.context['fields'][5]['options'], self.list_options)
        return

    #   -----------------------------------------------------------------------------
    #   Form rendering tests
    #   -----------------------------------------------------------------------------

    def test_get_form_rendering(self):
        u = entitydata_edit_url("new", "testcoll", "_type", view_id="Type_view")
        r = self.client.get(u+"?continuation_url=/xyzzy/")
        self.assertEqual(r.status_code,   200)
        self.assertEqual(r.reason_phrase, "OK")
        # log.info(r.content)
        self.assertContains(r, "<title>Collection testcoll</title>")
        self.assertContains(r, "<h3>'_type' data in collection 'testcoll'</h3>")
        field_vals = default_fields(coll_id="testcoll", type_id="_type", entity_id="00000001")
        formrow1 = """
            <div class="small-12 medium-6 columns">
              <div class="row view-value-row">
                <div class="%(label_classes)s">
                  <span>Id</span>
                </div>
                <div class="%(input_classes)s">
                  <input type="text" size="64" name="entity_id" 
                         placeholder="(type id)" value="00000001"/>
                </div>
              </div>
            </div>
            """%field_vals(width=6)
        formrow2 = """
            <div class="small-12 columns">
              <div class="row view-value-row">
                <div class="%(label_classes)s">
                  <span>Label</span>
                </div>
                <div class="%(input_classes)s">
                  <input type="text" size="64" name="Type_label" 
                         placeholder="(label)" 
                         value="%(default_label_esc)s" />
                </div>
              </div>
            </div>
            """%field_vals(width=12)
        formrow3 = """
            <div class="small-12 columns">
              <div class="row view-value-row">
                <div class="%(label_classes)s">
                  <span>Comment</span>
                </div>
                <div class="%(input_classes)s">
                  <textarea cols="64" rows="6" name="Type_comment" 
                            class="small-rows-4 medium-rows-8"
                            placeholder="(type description)"
                            >
                      %(default_comment_esc)s
                  </textarea>
                </div>
              </div>
            </div>
            """%field_vals(width=12)
        formrow4 = """
            <div class="small-12 columns">
              <div class="row view-value-row">
                <div class="%(label_classes)s">
                  <span>URI</span>
                </div>
                <div class="%(input_classes)s">
                  <input type="text" size="64" name="Type_uri" 
                         placeholder="(URI)"
                         value=""/>
                </div>
              </div>
            </div>
            """%field_vals(width=12)
        formrow5a = """
            <div class="%(space_classes)s">
              <div class="row">
                <div class="small-12 columns">
                  &nbsp;
                </div>
              </div>
            </div>
            """%field_vals(width=2)
        formrow5b = """
            <div class="%(button_wide_classes)s">
              <div class="row">
                <div class="%(button_left_classes)s">
                  <input type="submit" name="save"      value="Save" />
                  <input type="submit" name="cancel"    value="Cancel" />
                </div>
              </div>
            </div>
            """%field_vals(width=6)
        self.assertContains(r, formrow1, html=True)
        self.assertContains(r, formrow2, html=True)
        self.assertContains(r, formrow3, html=True)
        self.assertContains(r, formrow4, html=True)
        self.assertContains(r, formrow5a, html=True)
        self.assertContains(r, formrow5b, html=True)
        return

    def test_get_new(self):
        u = entitydata_edit_url("new", "testcoll", "_type", view_id="Type_view")
        r = self.client.get(u+"?continuation_url=/xyzzy/")
        self.assertEqual(r.status_code,   200)
        self.assertEqual(r.reason_phrase, "OK")
        # Test context
        type_url = collection_entity_view_url(coll_id="testcoll", type_id="_type", entity_id="00000001")
        self.assertEqual(r.context['coll_id'],          "testcoll")
        self.assertEqual(r.context['type_id'],          "_type")
        self.assertEqual(r.context['entity_id'],        "00000001")
        self.assertEqual(r.context['orig_id'],          "00000001")
        self.assertEqual(r.context['entity_url'],       type_url)
        self.assertEqual(r.context['entity_uri'],       None)
        self.assertEqual(r.context['action'],           "new")
        self.assertEqual(r.context['continuation_url'], "/xyzzy/")
        # Fields
        self._check_context_fields(r, 
            type_id="00000001",
            type_label=default_label("testcoll", "_type", "00000001"),
            type_help=default_comment("testcoll", "_type", "00000001"),
            type_uri=""
            )
        return

    def test_get_copy(self):
        u = entitydata_edit_url("copy", "testcoll", "_type", entity_id="Default_type", view_id="Type_view")
        r = self.client.get(u)
        self.assertEqual(r.status_code,   200)
        self.assertEqual(r.reason_phrase, "OK")
        # Test context (values read from test data fixture)
        type_url = collection_entity_view_url(coll_id="testcoll", type_id="_type", entity_id="Default_type")
        self.assertEqual(r.context['coll_id'],          "testcoll")
        self.assertEqual(r.context['type_id'],          "_type")
        self.assertEqual(r.context['entity_id'],        "Default_type")
        self.assertEqual(r.context['orig_id'],          "Default_type")
        self.assertEqual(r.context['entity_url'],       type_url)
        self.assertEqual(r.context['entity_uri'],       None)
        self.assertEqual(r.context['action'],           "copy")
        self.assertEqual(r.context['continuation_url'], "")
        # Fields
        self._check_context_fields(r, 
            type_id="Default_type",
            type_label="Default record type",
            type_help="Default record type, applied when no type is specified when creating a record.",
            type_uri=""
            )
        return

    def test_get_copy_not_exists(self):
        u = entitydata_edit_url("copy", "testcoll", "_type", entity_id="notype", view_id="Type_view")
        r = self.client.get(u)
        # log.info(r.content)
        self.assertEqual(r.status_code,   404)
        self.assertEqual(r.reason_phrase, "Not found")
        self.assertContains(r, "<title>Annalist error</title>", status_code=404)
        self.assertContains(r, "<h3>404: Not found</h3>", status_code=404)
        err_label = error_label("testcoll", "_type", "notype")
        self.assertContains(r, "<p>%s does not exist</p>"%(err_label), status_code=404)
        return

    def test_get_edit(self):
        u = entitydata_edit_url("edit", "testcoll", "_type", entity_id="Default_type", view_id="Type_view")
        r = self.client.get(u)
        self.assertEqual(r.status_code,   200)
        self.assertEqual(r.reason_phrase, "OK")
        # Test context (values read from test data fixture)
        type_url = collection_entity_view_url(coll_id="testcoll", type_id="_type", entity_id="Default_type")
        self.assertEqual(r.context['coll_id'],          "testcoll")
        self.assertEqual(r.context['type_id'],          "_type")
        self.assertEqual(r.context['entity_id'],        "Default_type")
        self.assertEqual(r.context['orig_id'],          "Default_type")
        self.assertEqual(r.context['entity_url'],       type_url)
        self.assertEqual(r.context['entity_uri'],       "annal:Default_type")
        self.assertEqual(r.context['action'],           "edit")
        self.assertEqual(r.context['continuation_url'], "")
        # Fields
        self._check_context_fields(r, 
            type_id="Default_type",
            type_label="Default record type",
            type_help="Default record type, applied when no type is specified when creating a record.",
            type_uri="annal:Default_type"
            )
        return

    def test_get_edit_not_exists(self):
        u = entitydata_edit_url("edit", "testcoll", "_type", entity_id="notype", view_id="Type_view")
        r = self.client.get(u)
        # log.info(r.content)
        self.assertEqual(r.status_code,   404)
        self.assertEqual(r.reason_phrase, "Not found")
        self.assertContains(r, "<title>Annalist error</title>", status_code=404)
        self.assertContains(r, "<h3>404: Not found</h3>", status_code=404)
        err_label = error_label("testcoll", "_type", "notype")
        self.assertContains(r, "<p>%s does not exist</p>"%(err_label), status_code=404)
        return

    #   -----------------------------------------------------------------------------
    #   Form response tests
    #   -----------------------------------------------------------------------------

    #   -------- new type --------

    def test_post_new_type(self):
        self.assertFalse(RecordType.exists(self.testcoll, "newtype"))
        f = recordtype_entity_view_form_data(type_id="newtype", action="new", update="RecordType")
        u = entitydata_edit_url("new", "testcoll", "_type", view_id="Type_view")
        r = self.client.post(u, f)
        # print r.content
        self.assertEqual(r.status_code,   302)
        self.assertEqual(r.reason_phrase, "FOUND")
        self.assertEqual(r.content,       "")
        self.assertEqual(r['location'], self.continuation_url)
        # Check that new record type exists
        self._check_record_type_values("newtype", update="RecordType")
        return

    def test_post_new_type_cancel(self):
        self.assertFalse(RecordType.exists(self.testcoll, "newtype"))
        f = recordtype_entity_view_form_data(
            type_id="newtype", action="new", cancel="Cancel", update="Updated RecordType"
            )
        u = entitydata_edit_url("new", "testcoll", "_type", view_id="Type_view")
        r = self.client.post(u, f)
        self.assertEqual(r.status_code,   302)
        self.assertEqual(r.reason_phrase, "FOUND")
        self.assertEqual(r.content,       "")
        self.assertEqual(r['location'], self.continuation_url)
        # Check that new record type still does not exist
        self.assertFalse(RecordType.exists(self.testcoll, "newtype"))
        return

    def test_post_new_type_missing_id(self):
        f = recordtype_entity_view_form_data(action="new", update="RecordType")
        u = entitydata_edit_url("new", "testcoll", "_type", view_id="Type_view")
        r = self.client.post(u, f)
        # print r.content
        self.assertEqual(r.status_code,   200)
        self.assertEqual(r.reason_phrase, "OK")
        self.assertContains(r, "<h3>Problem with record type identifier</h3>")
        # Test context
        expect_context = recordtype_entity_view_context_data(action="new", update="RecordType")
        self.assertDictionaryMatch(r.context, expect_context)
        return

    def test_post_new_type_invalid_id(self):
        f = recordtype_entity_view_form_data(
            type_id="!badtype", orig_id="orig_type_id", action="new", update="RecordType"
            )
        u = entitydata_edit_url("new", "testcoll", "_type", view_id="Type_view")
        r = self.client.post(u, f)
        self.assertEqual(r.status_code,   200)
        self.assertEqual(r.reason_phrase, "OK")
        self.assertContains(r, "<h3>Problem with record type identifier</h3>")
        # Test context
        expect_context = recordtype_entity_view_context_data(
            type_id="!badtype", orig_id="orig_type_id", action="new", update="RecordType"
            )
        self.assertDictionaryMatch(r.context, expect_context)
        return

    #   -------- copy type --------

    def test_post_copy_type(self):
        self.assertFalse(RecordType.exists(self.testcoll, "copytype"))
        f = recordtype_entity_view_form_data(
            type_id="copytype", orig_id="Default_type", action="copy", update="RecordType"
            )
        u = entitydata_edit_url("copy", "testcoll", "_type", entity_id="Default_type", view_id="Type_view")
        r = self.client.post(u, f)
        self.assertEqual(r.status_code,   302)
        self.assertEqual(r.reason_phrase, "FOUND")
        self.assertEqual(r.content,       "")
        self.assertEqual(r['location'], self.continuation_url)
        # Check that new record type exists
        self._check_record_type_values("copytype", update="RecordType")
        return

    def test_post_copy_type_cancel(self):
        self.assertFalse(RecordType.exists(self.testcoll, "copytype"))
        f = recordtype_entity_view_form_data(
            type_id="copytype", orig_id="Default_type", action="copy", cancel="Cancel", update="RecordType"
            )
        u = entitydata_edit_url("copy", "testcoll", "_type", entity_id="Default_type", view_id="Type_view")
        r = self.client.post(u, f)
        self.assertEqual(r.status_code,   302)
        self.assertEqual(r.reason_phrase, "FOUND")
        self.assertEqual(r.content,       "")
        self.assertEqual(r['location'], self.continuation_url)
        # Check that target record type still does not exist
        self.assertFalse(RecordType.exists(self.testcoll, "copytype"))
        return

    def test_post_copy_type_missing_id(self):
        f = recordtype_entity_view_form_data(
            action="copy", update="Updated RecordType"
            )
        u = entitydata_edit_url("copy", "testcoll", "_type", entity_id="Default_type", view_id="Type_view")
        r = self.client.post(u, f)
        self.assertEqual(r.status_code,   200)
        self.assertEqual(r.reason_phrase, "OK")
        self.assertContains(r, "<h3>Problem with record type identifier</h3>")
        expect_context = recordtype_entity_view_context_data(action="copy", update="Updated RecordType")
        self.assertDictionaryMatch(r.context, expect_context)
        return

    def test_post_copy_type_invalid_id(self):
        f = recordtype_entity_view_form_data(
            type_id="!badtype", orig_id="Default_type", action="copy", update="Updated RecordType"
            )
        u = entitydata_edit_url("copy", "testcoll", "_type", entity_id="Default_type", view_id="Type_view")
        r = self.client.post(u, f)
        self.assertEqual(r.status_code,   200)
        self.assertEqual(r.reason_phrase, "OK")
        self.assertContains(r, "<h3>Problem with record type identifier</h3>")
        expect_context = recordtype_entity_view_context_data(
            type_id="!badtype", orig_id="Default_type", 
            action="copy", update="Updated RecordType"
            )
        self.assertDictionaryMatch(r.context, expect_context)
        return

    #   -------- edit type --------

    def test_post_edit_type(self):
        self._create_record_type("edittype")
        self._check_record_type_values("edittype")
        f = recordtype_entity_view_form_data(
            type_id="edittype", orig_id="edittype", 
            action="edit", update="Updated RecordType"
            )
        u = entitydata_edit_url("edit", "testcoll", "_type", entity_id="edittype", view_id="Type_view")
        r = self.client.post(u, f)
        self.assertEqual(r.status_code,   302)
        self.assertEqual(r.reason_phrase, "FOUND")
        self.assertEqual(r.content,       "")
        self.assertEqual(r['location'], self.continuation_url)
        # Check that new record type exists
        self._check_record_type_values("edittype", update="Updated RecordType")
        return

    def test_post_edit_type_new_id(self):
        # Check logic applied when type is renamed
        (t, d1, e1) = self._create_record_type("edittype1", entity_id="typeentity")
        self.assertTrue(RecordType.exists(self.testcoll, "edittype1"))
        self.assertFalse(RecordType.exists(self.testcoll, "edittype2"))
        self.assertTrue(RecordTypeData.exists(self.testcoll, "edittype1"))
        self.assertFalse(RecordTypeData.exists(self.testcoll, "edittype2"))
        self.assertTrue(EntityData.exists(d1, "typeentity"))
        self._check_record_type_values("edittype1")
        f = recordtype_entity_view_form_data(
            type_id="edittype2", orig_id="edittype1", 
            action="edit", update="Updated RecordType"
            )
        u = entitydata_edit_url("edit", "testcoll", "_type", entity_id="edittype1", view_id="Type_view")
        r = self.client.post(u, f)
        self.assertEqual(r.status_code,   302)
        self.assertEqual(r.reason_phrase, "FOUND")
        self.assertEqual(r.content,       "")
        self.assertEqual(r['location'], self.continuation_url)
        # Check that new record type exists and old does not
        self.assertFalse(RecordType.exists(self.testcoll, "edittype1"))
        self.assertTrue(RecordType.exists(self.testcoll, "edittype2"))
        self._check_record_type_values("edittype2", update="Updated RecordType")
        # Check that type data directory has been renamed
        self.assertFalse(RecordTypeData.exists(self.testcoll, "edittype1"))
        self.assertTrue(RecordTypeData.exists(self.testcoll, "edittype2"))
        self.assertFalse(EntityData.exists(d1, "typeentity"))
        d2 = RecordTypeData.load(self.testcoll, "edittype2")
        self.assertTrue(EntityData.exists(d2, "typeentity"))
        return

    def test_post_edit_type_cancel(self):
        self._create_record_type("edittype")
        self._check_record_type_values("edittype")
        f = recordtype_entity_view_form_data(
            type_id="edittype", orig_id="edittype", 
            action="edit", cancel="Cancel", update="Updated RecordType"
            )
        u = entitydata_edit_url("edit", "testcoll", "_type", entity_id="edittype", view_id="Type_view")
        r = self.client.post(u, f)
        self.assertEqual(r.status_code,   302)
        self.assertEqual(r.reason_phrase, "FOUND")
        self.assertEqual(r.content,       "")
        self.assertEqual(r['location'], self.continuation_url)
        # Check that target record type still does not exist and unchanged
        self._check_record_type_values("edittype")
        return

    def test_post_edit_type_missing_id(self):
        self._create_record_type("edittype")
        self._check_record_type_values("edittype")
        # Form post with ID missing
        f = recordtype_entity_view_form_data(
            action="edit", update="Updated RecordType"
            )
        u = entitydata_edit_url("edit", "testcoll", "_type", entity_id="edittype", view_id="Type_view")
        r = self.client.post(u, f)
        self.assertEqual(r.status_code,   200)
        self.assertEqual(r.reason_phrase, "OK")
        self.assertContains(r, "<h3>Problem with record type identifier</h3>")
        # Test context for re-rendered form
        expect_context = recordtype_entity_view_context_data(action="edit", update="Updated RecordType")
        self.assertDictionaryMatch(r.context, expect_context)
        # Check original data is unchanged
        self._check_record_type_values("edittype")
        return

    def test_post_edit_type_invalid_id(self):
        self._create_record_type("edittype")
        self._check_record_type_values("edittype")
        # Form post with invalid ID
        f = recordtype_entity_view_form_data(
            type_id="!badtype", orig_id="edittype", action="edit", update="Updated RecordType"
            )
        u = entitydata_edit_url("edit", "testcoll", "_type", entity_id="edittype", view_id="Type_view")
        r = self.client.post(u, f)
        self.assertEqual(r.status_code,   200)
        self.assertEqual(r.reason_phrase, "OK")
        self.assertContains(r, "<h3>Problem with record type identifier</h3>")
        # Test context
        expect_context = recordtype_entity_view_context_data(
            type_id="!badtype", orig_id="edittype", 
            action="edit", update="Updated RecordType"
            )
        self.assertDictionaryMatch(r.context, expect_context)
        # Check original data is unchanged
        self._check_record_type_values("edittype")
        return

#   -----------------------------------------------------------------------------
#
#   ConfirmRecordTypeDeleteTests tests for completion of record deletion
#
#   -----------------------------------------------------------------------------

class ConfirmRecordTypeDeleteTests(AnnalistTestCase):
    """
    Tests for record type deletion on response to confirmation form
    """

    def setUp(self):
        init_annalist_test_site()
        self.testsite = Site(TestBaseUri, TestBaseDir)
        self.testcoll = Collection.create(self.testsite, "testcoll", collection_create_values("testcoll"))
        # Login and permissions
        create_test_user(self.testcoll, "testuser", "testpassword")
        self.client = Client(HTTP_HOST=TestHost)
        loggedin = self.client.login(username="testuser", password="testpassword")
        self.assertTrue(loggedin)
        return

    def tearDown(self):
        return

    def test_CollectionActionViewTest(self):
        self.assertEqual(RecordTypeDeleteConfirmedView.__name__, "RecordTypeDeleteConfirmedView", "Check RecordTypeDeleteConfirmedView class name")
        return

    # NOTE:  test_collection checks the appropriate response from clicking the delete button, 
    # so here only need to test completion code.
    def test_post_confirmed_remove_type(self):
        t = RecordType.create(self.testcoll, "deletetype", recordtype_create_values("deletetype"))
        self.assertTrue(RecordType.exists(self.testcoll, "deletetype"))
        # Submit positive confirmation
        u = TestHostUri + recordtype_edit_url("delete", "testcoll")
        f = recordtype_delete_confirm_form_data("deletetype")
        r = self.client.post(u, f)
        self.assertEqual(r.status_code,     302)
        self.assertEqual(r.reason_phrase,   "FOUND")
        self.assertEqual(r.content,         "")
        self.assertMatch(r['location'],    
            "^"+TestHostUri+
            collection_edit_url("testcoll")+
            r"\?info_head=.*&info_message=.*deletetype.*testcoll.*$"
            )
        # Confirm deletion
        self.assertFalse(RecordType.exists(self.testcoll, "deletetype"))
        return

# End.



"""
Annalist record field description
"""

__author__      = "Graham Klyne (GK@ACM.ORG)"
__copyright__   = "Copyright 2014, G. Klyne"
__license__     = "MIT (http://opensource.org/licenses/MIT)"

import os
import os.path
import urlparse
import shutil

import logging
log = logging.getLogger(__name__)

from django.conf import settings

from annalist                   import layout
from annalist.identifiers       import ANNAL
from annalist.models.entity     import Entity
from annalist.models.entitydata import EntityData

class RecordField(EntityData):

    _entitytype     = ANNAL.CURIE.Field
    _entitytypeid   = "_field"
    _entityview     = layout.COLL_FIELD_VIEW
    _entitypath     = layout.COLL_FIELD_PATH
    _entityaltpath  = layout.SITE_FIELD_PATH
    _entityfile     = layout.FIELD_META_FILE
    _entityref      = layout.META_FIELD_REF

    def __init__(self, parent, field_id, altparent=None):
        """
        Initialize a new RecordField object, without metadta (yet).

        parent      is the parent entity from which the view is descended.
        field_id    the local identifier for the record view
        altparent   is a site object to search for this new entity,
                    allowing site-wide RecordField values to be found.
        """
        log.debug("RecordField %s"%(field_id))
        # assert altparent, "RecordField instantiated with no altparent"
        super(RecordField, self).__init__(parent, field_id, altparent=altparent)
        return

# End.



"""
Various utilities for Annalist common functions
"""

__author__      = "Graham Klyne (GK@ACM.ORG)"
__copyright__   = "Copyright 2014, G. Klyne"
__license__     = "MIT (http://opensource.org/licenses/MIT)"

import logging
log = logging.getLogger(__name__)

import re
import os
import os.path
import errno
import stat
import time
import urlparse
import json
import shutil
import StringIO

from django.conf import settings

from annalist.identifiers import ANNAL

def valid_id(id):
    """
    Checks the supplied id is valid as an Annalist identifier.

    The main requirement is that it is valid as a URI path segment, so it can be used
    in the creation of URIs for Annalist resources.  Also, filters out reserved Annalist
    identifiers.

    >>> valid_id("abcdef_1234")
    True
    >>> valid_id("abcdef/1234")
    False
    >>> valid_id("_annalist_collection")
    False
    """
    reserved = (
        [ "_annalist_site"
        , "_annalist_collection"
        ])
    # cf. urls.py:
    if id and re.match(r"\w{1,32}$", id):
        return id not in reserved
    return False

def slug_from_name(filename):
    """
    Extracts a slug (id) value from a filename

    >>> slug_from_path("bar.baz")
    'bar'
    >>> slug_from_path("bar")
    'bar'
    >>> slug_from_path(".baz")
    '.baz'
    """
    slug = os.path.splitext(filename)[0]
    return slug

def slug_from_path(path):
    """
    Extracts a slug (id) value from a file path

    >>> slug_from_path("/foo/bar.baz")
    'bar'
    >>> slug_from_path("/bar")
    'bar'
    >>> slug_from_path("bar")
    'bar'
    >>> slug_from_path("/example.org/foo/bar/.baz")
    '.baz'
    >>> slug_from_path("/foo/bar.baz")
    'bar'
    >>> slug_from_path("/example.org/foo/bar/")+"$"
    '$'
    """
    return slug_from_name(os.path.basename(path))

def slug_from_uri(uri):
    """
    Extracts a slug (id) value from a URI

    >>> slug_from_uri("http:/example.org/foo/bar")
    'bar'
    >>> slug_from_uri("/example.org/foo/bar")
    'bar'
    >>> slug_from_uri("/foo/bar")
    'bar'
    >>> slug_from_uri("/bar")
    'bar'
    >>> slug_from_uri("bar")
    'bar'
    >>> slug_from_uri("/example.org/foo/bar/")+"$"
    '$'
    >>> slug_from_uri("http:/example.org/foo/bar.baz")
    'bar'
    >>> slug_from_uri("http:/example.org/foo/bar;baz")
    'bar;baz'
    >>> slug_from_uri("http:/example.org/foo/bar?baz")
    'bar'
    >>> slug_from_uri("http:/example.org/foo/bar#baz")
    'bar'
    """
    return slug_from_path(urlparse.urlsplit(uri).path)

def ensure_dir(dirname):
    """
    Ensure that a named directory exists; if it does not, attempt to create it.
    """
    try:
        os.makedirs(dirname)
    except OSError, e:
        if e.errno != errno.EEXIST:
            raise
    return

def entity_dir_path(base_dir, path, filename):
    """
    Assemble full entity description directory and file names from supplied components.

    base_dir    is the fully qualified base directory for the site or collection
                from which data is to be read
    path        is a relative path or list of path segments within the site or 
                resource for the site or collection from which data is to be read.
                This value may be absent.
    filename    is a file name for the data resource to be read.

    Returns a pair containing the full directory and path names for the entity file.

    >>> entity_dir_path("/base/dir/","sub","file.ext")
    ('/base/dir/sub', '/base/dir/sub/file.ext')
    >>> entity_dir_path("/base/dir/",["sub"],"file.ext")
    ('/base/dir/sub', '/base/dir/sub/file.ext')
    >>> entity_dir_path("/base/dir/",[],"file.ext")
    ('/base/dir', '/base/dir/file.ext')
    >>> entity_dir_path("/base/dir/",["sub1","sub2"],"file.ext")
    ('/base/dir/sub1/sub2', '/base/dir/sub1/sub2/file.ext')
    >>> entity_dir_path("/base/dir",["sub"],"file.ext")
    ('/base/dir/sub', '/base/dir/sub/file.ext')
    >>> entity_dir_path("/base/dir",[],"sub/file.ext")
    ('/base/dir/sub', '/base/dir/sub/file.ext')
    """
    log.debug("entity_dir_path %s, %r, %s"%(base_dir, path, filename))
    if path:
        if isinstance(path, (list, tuple)):
            d = os.path.join(base_dir, *path)
        else:
            d = os.path.join(base_dir, path)
    else:
        d = base_dir
    p = os.path.join(d, filename)
    d = os.path.dirname(p)
    return (d, p)

def entity_path(base_dir, path, filename):
    """
    Assemble full entity description file names from supplied components.

    base_dir    is the fully qualified base directory for the site or collection
                from which data is to be read
    path        is a relative path or list of path segments within the site or 
                resource for the site or collection from which data is to be read.
                This value may be absent.
    filename    is a file name for the data resource to be read.

    Returns the full path name for the entity file if the intermediate directories 
    all exist, otherwise None if any directories are missing.  No test is made for 
    existence of the filename.

    >>> entity_path('.',[],"file.ext")
    './file.ext'
    >>> entity_path('.',["nopath"],"file.ext") is None
    True
    """
    (d, p) = entity_dir_path(base_dir, path, filename)
    # log.debug("entity_path: d %s, p %s"%(d,p))
    if d and os.path.isdir(d):
        return p
    return None

def entity_url_host(baseuri, entityref):
    """
    Return host part (as appears in an HTTP host: header) from an entity URI.

    >>> entity_url_host("http://example.org/basepath/", "/path/to/entity")
    'example.org'
    >>> entity_url_host("http://example.org:80/basepath/", "/path/to/entity")
    'example.org:80'
    >>> entity_url_host("http://userinfo@example.org:80/basepath/", "/path/to/entity")
    'example.org:80'
    >>> entity_url_host("http://base.example.org:80/basepath/", "http://ref.example.org/path/to/entity")
    'ref.example.org'
    """
    uri = urlparse.urljoin(baseuri, entityref)
    p   = urlparse.urlparse(uri)
    h   = p.hostname or ""
    if p.port:
        h += ":" + str(p.port)
    return h

    """
    Return absolute path part from an entity URI, excluding query or fragment.

    >>> entity_url_host("http://example.org/basepath/", "/path/to/entity")
    '/path/to/entity'
    >>> entity_url_host("http://example.org/basepath/", "relpath/to/entity")
    '/basepath/relpath/to/entity'
    >>> entity_url_host("http://example.org/basepath/", "/path/to/entity?query")
    '/path/to/entity'
    >>> entity_url_host("http://example.org/basepath/", "/path/to/entity#frag")
    '/path/to/entity'
    >>> entity_url_host("/basepath/", "relpath/to/entity")
    '/basepath/relpath/to/entity'
    """
def entity_url_path(baseuri, entityref):
    uri = urlparse.urljoin(baseuri, entityref)
    return urlparse.urlparse(uri).path

def strip_comments(f):
    """
    Returns a file-like object that returns content from the supplied file-like
    but with comments removed.

    >>> f1 = StringIO.StringIO("// coment\\ndata\\n// another comment\\n\\n")
    >>> f2 = strip_comments(f1)
    >>> f2.read()
    '\\ndata\\n\\n\\n'
    """
    fnc = StringIO.StringIO()
    sof = fnc.tell()
    for line in f:
        if re.match("^\s*//", line):
            fnc.write("\n")
        else:
            fnc.write(line)
    fnc.seek(sof)
    return fnc

def renametree_temp(src):
    """
    Rename tree to temporary name, and return that name, or 
    None if the source directory does not exist.
    """
    count = 0 
    while count < 10:      # prevents indefinite loop
        count += 1
        tmp = os.path.join(os.path.dirname(src),"_removetree_tmp_%d"%(count))
        try:
            os.rename(src, tmp)
            return tmp      # Success!
        # except WindowsError as e:
        #     log.warning(
        #         "util.renametree_temp: WindowsError: winerror %d, strerror %s, errno %d"%
        #         (e.winerror, e.strerror, e.errno)
        #         )
        #     continue        # Try another temp name
        except OSError as e:
            time.sleep(1)
            if e.errno == errno.EACCES:
                log.warning("util.renametree_temp: %s EACCES, retrying"%tmp)
                continue    # Try another temp name
            if e.errno == errno.ENOTEMPTY:
                log.warning("util.renametree_temp: %s ENOTEMPTY, retrying"%tmp)
                continue    # Try another temp name
            if e.errno == errno.EEXIST:
                log.warning("util.renametree_temp: %s EEXIST, retrying"%tmp)
                shutil.rmtree(tmp, ignore_errors=True)  # Try to clean up old files
                continue    # Try another temp name
            if e.errno == errno.ENOENT:
                log.warning("util.renametree_temp: %s ENOENT, skipping"%tmp)
                break       # 'src' does not exist(?)
            raise           # Other error: propagaee
    return None

def removetree(tgt):
    """
    Work-around for python problem with shutils tree remove functions on Windows.
    See:
        http://stackoverflow.com/questions/23924223/
        http://stackoverflow.com/questions/1213706/
        http://stackoverflow.com/questions/1889597/
        http://bugs.python.org/issue19643
    """
    # shutil.rmtree error handler that attempts recovery on Windows from 
    # attempts to remove a read-only file or directory (see links above).
    def error_handler(func, path, execinfo):
        # figure out recovery based on error...
        e = execinfo[1]
        if e.errno == errno.ENOENT or not os.path.exists(path):
            return          # path does not exist
        if func in (os.rmdir, os.remove) and e.errno == errno.EACCES:
            try:
                os.chmod(path, stat.S_IRWXU| stat.S_IRWXG| stat.S_IRWXO) # 0777
            except Exception as che:
                log.warning("util.removetree: chmod failed: %s"%che)
            try:
                func(path)
            except Exception as rfe:
                log.warning("util.removetree: 'func' retry failed: %s"%rfe)
                if not os.path.exists(path):
                    return      # Gone, assume all is well
                raise
        if e.errno == errno.ENOTEMPTY:
            log.warning("util.removetree: Not empty: %s, %s"%(path, tgt))
            time.sleep(1)
            removetree(path)    # Retry complete removal
            return
        log.warning("util.removetree: rmtree path: %s, error: %s"%(path, repr(execinfo)))
        raise e
    # Workaround for problems on Windows: it appears that the directory
    # removal does not complete immediately, causing subsequent failures.
    # Try renaming to a new directory first, so that the tgt is immediately 
    # available for re-use.
    tmp = renametree_temp(tgt)
    if tmp:
        shutil.rmtree(tmp, onerror=error_handler)
    return

def replacetree(src, tgt):
    """
    Work-around for python problem with shutils tree copy functions on Windows.
    See: http://stackoverflow.com/questions/23924223/
    """
    if os.path.exists(tgt):
        removetree(tgt)
    shutil.copytree(src, tgt)
    return

def updatetree(src, tgt):
    """
    Like replacetree, except that existing files are not removed unless replaced 
    by a file of the name name in the source tree.

    NOTE: can't use shutil.copytree for this, as that requires that the 
    destination tree does not exist.
    """
    files = os.listdir(src)
    for f in files:
        sf = os.path.join(src, f)
        if os.path.exists(sf) and not os.path.islink(sf):   # Ignore symlinks
            if os.path.isdir(sf):
                tf = os.path.join(tgt, f)
                if not os.path.isdir(tf):
                    os.makedirs(tf)
                updatetree(sf, tf)                          # Recursive dir copy
            else:
                shutil.copy2(sf, tgt)                       # Copy single file, may overwrite
    return

if __name__ == "__main__":
    import doctest
    doctest.testmod()

# End.



"""
Define various message strings generated in the code
"""

__author__      = "Graham Klyne (GK@ACM.ORG)"
__copyright__   = "Copyright 2014, G. Klyne"
__license__     = "MIT (http://opensource.org/licenses/MIT)"

import logging
log = logging.getLogger(__name__)

SITE_NAME_DEFAULT           = "Annalist linked data notebook"
ACTION_COMPLETED            = "Action completed"
NO_ACTION_PERFORMED         = "No action performed"
INPUT_ERROR                 = "Problem with input"
SYSTEM_ERROR                = "System error"
UNEXPECTED_FORM_DATA        = "Unexpected form data: %r"
MISSING_COLLECTION_ID       = "Missing identifier for new collection"
INVALID_COLLECTION_ID       = "Invalid identifier for new collection: '%(coll_id)s'"
CREATED_COLLECTION_ID       = "Created new collection: '%(coll_id)s'"
REMOVE_COLLECTIONS          = "Remove collection(s): %(ids)s"
NO_COLLECTIONS_SELECTED     = "No collections selected for removal"
COLLECTIONS_REMOVED         = "The following collections were removed: %(ids)s"
NO_TYPE_FOR_COPY            = "No record type selected to copy"
NO_TYPE_FOR_EDIT            = "No record type selected to edit"
NO_TYPE_FOR_DELETE          = "No record type selected to delete"
NO_VIEW_FOR_COPY            = "No record view selected to copy"
NO_VIEW_FOR_EDIT            = "No record view selected to edit"
NO_VIEW_FOR_DELETE          = "No record view selected to delete"
NO_LIST_FOR_COPY            = "No list view selected to copy"
NO_LIST_FOR_EDIT            = "No list view selected to edit"
NO_LIST_FOR_DELETE          = "No list view selected to delete"
DOES_NOT_EXIST              = "%(id)s does not exist"
REMOVE_RECORD_TYPE          = "Remove record type %(id)s in collection %(coll_id)s"
REMOVE_RECORD_LIST          = "Remove record list %(id)s in collection %(coll_id)s"
REMOVE_RECORD_VIEW          = "Remove record view %(id)s in collection %(coll_id)s"
COLLECTION_ID               = "Problem with collection identifier"
COLLECTION_ID_INVALID       = "The collection identifier is missing or not a valid identifier"
COLLECTION_LABEL            = "Collection %(id)s"
COLLECTION_EXISTS           = "Collection %(id)s already exists"
COLLECTION_NOT_EXISTS       = "Collection %(id)s does not exist"
ANNALIST_USER_ID            = "Problem with user identifier"
ANNALIST_USER_ID_INVALID    = "The user identifier is missing or not a valid identifier"
ANNALIST_USER_LABEL         = "User %(id)s in collection %(coll_id)s"
ANNALIST_USER_EXISTS        = "User %(id)s in collection %(coll_id)s already exists"
ANNALIST_USER_NOT_EXISTS    = "User %(id)s in collection %(coll_id)s does not exist"
ANNALIST_USER_REMOVED       = "User %(id)s in collection %(coll_id)s was removed"
RECORD_TYPE_ID              = "Problem with record type identifier"
RECORD_TYPE_ID_INVALID      = "The record type identifier is missing or not a valid identifier"
RECORD_TYPE_LABEL           = "Record type %(id)s in collection %(coll_id)s"
RECORD_TYPE_EXISTS          = "Record type %(id)s in collection %(coll_id)s already exists"
RECORD_TYPE_NOT_EXISTS      = "Record type %(id)s in collection %(coll_id)s does not exist"
RECORD_TYPE_REMOVED         = "Record type %(id)s in collection %(coll_id)s was removed"
RECORD_LIST_ID              = "Problem with record list identifier"
RECORD_LIST_ID_INVALID      = "The record list identifier is missing or not a valid identifier"
RECORD_LIST_LABEL           = "Record list %(id)s in collection %(coll_id)s"
RECORD_LIST_EXISTS          = "Record list %(id)s in collection %(coll_id)s already exists"
RECORD_LIST_NOT_EXISTS      = "Record list %(id)s in collection %(coll_id)s does not exist"
RECORD_LIST_REMOVED         = "Record list %(id)s in collection %(coll_id)s was removed"
RECORD_VIEW_ID              = "Problem with record view identifier"
RECORD_VIEW_ID_INVALID      = "The record view identifier is missing or not a valid identifier"
RECORD_VIEW_LABEL           = "Record view %(id)s in collection %(coll_id)s"
RECORD_VIEW_EXISTS          = "Record view %(id)s in collection %(coll_id)s already exists"
RECORD_VIEW_NOT_EXISTS      = "Record view %(id)s in collection %(coll_id)s does not exist"
RECORD_VIEW_REMOVED         = "Record view %(id)s in collection %(coll_id)s was removed"
RECORD_GROUP_ID             = "Problem with field group identifier"
RECORD_GROUP_ID_INVALID     = "The field group identifier is missing or not a valid identifier"
RECORD_GROUP_LABEL          = "Field group %(id)s in collection %(coll_id)s"
RECORD_GROUP_EXISTS         = "Field group %(id)s in collection %(coll_id)s already exists"
RECORD_GROUP_NOT_EXISTS     = "Field group %(id)s in collection %(coll_id)s does not exist"
RECORD_GROUP_REMOVED        = "Field group %(id)s in collection %(coll_id)s was removed"
RECORD_FIELD_ID             = "Problem with record field identifier"
RECORD_FIELD_ID_INVALID     = "The record field identifier is missing or not a valid identifier"
RECORD_FIELD_LABEL          = "Record field %(id)s in collection %(coll_id)s"
RECORD_FIELD_EXISTS         = "Record field %(id)s in collection %(coll_id)s already exists"
RECORD_FIELD_NOT_EXISTS     = "Record field %(id)s in collection %(coll_id)s does not exist"
RECORD_FIELD_REMOVED        = "Record field %(id)s in collection %(coll_id)s was removed"
RECORD_ENUM_ID              = "Problem with enumeration type identifier"
RECORD_ENUM_ID_INVALID      = "The enumeration type identifier is missing or not a valid identifier"
RECORD_ENUM_LABEL           = "Enumeration type %(id)s in collection %(coll_id)s"
RECORD_ENUM_EXISTS          = "Enumeration type %(id)s in collection %(coll_id)s already exists"
RECORD_ENUM_NOT_EXISTS      = "Enumeration type %(id)s in collection %(coll_id)s does not exist"
RECORD_ENUM_REMOVED         = "Enumeration type %(id)s in collection %(coll_id)s was removed"
TOO_MANY_ENTITIES_SEL       = "Too many items selected"
NO_ENTITY_FOR_COPY          = "No data record selected to copy"
NO_ENTITY_FOR_EDIT          = "No data record selected to edit"
NO_ENTITY_FOR_DELETE        = "No data record selected to delete"
SITE_ENTITY_FOR_DELETE      = "Cannot remove site built-in entity %(id)s, or entity not found"
TYPE_VALUES_FOR_DELETE      = "Cannot remove type %(type_id)s with existing values"
REMOVE_ENTITY_DATA          = "Remove record %(id)s of type %(type_id)s in collection %(coll_id)s"
ENTITY_DATA_ID              = "Problem with entity identifier"
ENTITY_DATA_ID_INVALID      = "The entity identifier is missing, too long, or not a valid identifier"
ENTITY_DATA_LABEL           = "Entity %(id)s of type %(type_id)s in collection %(coll_id)s"
ENTITY_DATA_EXISTS          = "Entity %(id)s of type %(type_id)s in collection %(coll_id)s already exists"
ENTITY_DATA_NOT_EXISTS      = "Entity %(id)s of type %(type_id)s in collection %(coll_id)s does not exist"
ENTITY_DATA_REMOVED         = "Entity %(id)s of type %(type_id)s in collection %(coll_id)s was removed"
ENTITY_TYPE_ID              = "Problem with entity type identifier"
ENTITY_TYPE_ID_INVALID      = "The entity type identifier is missing, too long, or not a valid identifier"
DEFAULT_VIEW_UPDATED        = "Default list view for collection %(coll_id)s changed to %(list_id)s"
REMOVE_FIELD_ERROR          = "Problem with remove field(s) request"
NO_FIELD_SELECTED           = "No field(s) selected"
ENTITY_MESSAGE_LABEL        = "Entity %(type_id)s/%(entity_id)s in collection %(coll_id)s"
ENTITY_DEFAULT_LABEL        = ""    # "Entity %(type_id)s/%(entity_id)s in collection %(coll_id)s"
ENTITY_DEFAULT_COMMENT      = ""    # "Entity %(type_id)s/%(entity_id)s in collection %(coll_id)s"
VIEW_DESCRIPTION_HEADING    = "Problem with view description"
VIEW_PROPERTY_DUPLICATE     = "Field %(field_id)s repeats use of property %(property_uri)s in view"
INVALID_OPERATION_ATTEMPTED = "Attempt to peform invalid operation"
INVALID_TYPE_CHANGE         = "Change of record type to or from '_type' is not supported"
INVALID_TYPE_RENAME         = "Renaming of Annalist built-in types is not supported"
CREATE_ENTITY_FAILED        = "Problem creating/updating entity %s/%s (see log for more info)"
RENAME_ENTITY_FAILED        = "Problem renaming entity %s/%s to %s/%s (see log for more info)"
RENAME_TYPE_FAILED          = "Problem renaming type %s to %s (see log for more info)"

# End.



"""
Annalist application
"""

__author__        = "Graham Klyne"
__author_email__  = "GK@ACM.ORG"
__copyright__     = "Copyright 2014, G. Klyne"
__license__       = "MIT (http://opensource.org/licenses/MIT)"

# https://docs.djangoproject.com/en/dev/ref/applications/
default_app_config = 'annalist.apps.AnnalistConfig'

# Version 0.1: initial public prototype
#__version__        = "0.1.1"   # Feature freeze
#__version__        = "0.1.2"   # Test with Django 1.7
                                # Initial installation kit
                                # Apply sorting to entity lists to make test cases more robust across systems
                                # Setup scripts to initialize installation and site data
                                # Fix problem with local logins
                                # Various minor page presentation changes
                                # Access root URI path ('/') redirects to site display
                                # Add some online help text for site, collection front pages, and login page
                                # Resolve virtualenv problem on Ubuntu 14.04
                                # util.removetree hack to allow test suite to run on Windows
                                # Initial documentation
                                # Demonstration screencast
#__version__        = "0.1.3"   # Bump version number
                                # Post-release documentation updates
                                # Update annalist_manager with updatesitedata command and 
                                #   to create auth providers directory
                                # Updated field view definition to include extra fields used for 
                                #   enumerated type displays, etc.
                                # Updated view editing form description.
                                # Update field and view description of fields to restrict 
                                #   presentation of field choices
                                # Fix problem of ignoring blank value in submitted form (issue #30)
                                # Use collection and list labels for headings on entity list page (issue #26)
                                # Clean up page and section headings in record editing view
                                # Change confusing 'Select' label of field id (Field_sel) dropdown in view description
                                # Initialize entity label and comment to blank (issue #24)
                                # Fixed problem with rename locally created Default_view (issue #22)
                                # Fix that changing type of entity was not deleting old record (issue #29)
# __version__        = "0.1.4"  # Bump version number
# __version__        = "0.1.5"  # Bump version number (odd = unstable)
                                # Implemented authorization framework
                                # More new commands in `annalist-manager`
                                # List view option to hide columns on smaller screens
                                # Extend test suite covereage
                                # When adding field to view: check property URI is unique
                                # Documentation updates and new screencasts
                                # When renaming type, rename insrances to new type
                                # Prevent deleting type with instances present
                                # If specified default list not found, revert to built-in default
                                # Fix bug in entity reference field links (was linking to self, not target record)
                                # Add optional enumeration to available field render types
                                # Use separate Django database for each configuration
                                # Improve log file handling; include timestamps
                                # Some small usability improvements
                                # Address some areas of technical debt
#__version__        = "0.1.6"   # Bump version number for release
#__version__        = "0.1.6a"  # Hotfix: server error copying view without URI field
#__version__        = "0.1.7"   # Bump version number (odd = unstable)
                                # Rework form generator; full support for repeating fields
                                # Enlarge test coverage
                                # Some presentation improvements
                                # Various smaller bug fixes
#__version__        = "0.1.8"   # Bump version number for release
#__version__        = "0.1.8a"  # Hotfix: https://github.com/gklyne/annalist/issues/39 (bad username causes server error)
#__version__        = "0.1.9"   # Bump version number (odd = unstable)
                                # Add field alias option
                                # Option ("+") to create new linked record from referring record view
                                # Rework handling of repeated use of same field
                                # When changing list view, don't filter by entity type
                                # Various bug fixes
                                # Form display and response code refactoring
                                # Use placholder as descriptive label foprn repeat group form display
                                # Refactor field rendering logic
                                # Implement special renderer for field placement (position/size)
                                # Document default admin user permissions created with new site
                                # annalist-manager enhancements; permissions help, new subcommands
                                # Added build scripts for Docker containers
#__version__        = "0.1.10"  # Bump version number for release
#__version__        = "0.1.11"  # Bump version number (odd = unstable)
                                # Save logs in root of annalist_site data, for Docker visibility
                                # Introduce non-edit view of entities, with navigation links
                                # New render type: Boolean as checkbox
                                # New render type: URI as Hyperlink
                                # New render type: URI as embedded image
                                # New render type: long text with Markdown formatting; and new CSS
                                # Page layout/styling changes; rationalize some CSS usage
                                # Change 'Add field' button to 'Edit view'
                                # Add 'View description' button on non-editing entity views
                                # View display: suppress headings for empty repeatgrouprow value
                                # Preserve current value in form if not present in drop-down options
                                # Various bug fixes (see release notes)
__version__        = "0.1.12"  # Bump version number for release


# End.
